'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _without = _interopDefault(require('lodash/without'));
var _forEach = _interopDefault(require('lodash/forEach'));
var _isArray = _interopDefault(require('lodash/isArray'));
var _toArray = _interopDefault(require('lodash/toArray'));
var matchesSelector = _interopDefault(require('dom-matches'));
var Trackr = _interopDefault(require('trackr'));
var _has = _interopDefault(require('lodash/has'));
var events = require('events');
var trackrObjects = require('trackr-objects');
var assignProps = _interopDefault(require('assign-props'));
var _assign = _interopDefault(require('lodash/assign'));
var _includes = _interopDefault(require('lodash/includes'));
var _isFunction = _interopDefault(require('lodash/isFunction'));
var _isString = _interopDefault(require('lodash/isString'));
var raf = _interopDefault(require('raf'));
var _isObject = _interopDefault(require('lodash/isObject'));
var _defer = _interopDefault(require('lodash/defer'));
var _pick = _interopDefault(require('lodash/pick'));
var _isDate = _interopDefault(require('lodash/isDate'));
var subclass = _interopDefault(require('backbone-extend-standalone'));
var sourceMap = require('source-map');
var _uniqueId = _interopDefault(require('lodash/uniqueId'));
var _invokeMap = _interopDefault(require('lodash/invokeMap'));
var path = _interopDefault(require('path'));
var _last = _interopDefault(require('lodash/last'));
var _map = _interopDefault(require('lodash/map'));
var _find = _interopDefault(require('lodash/find'));
var _union = _interopDefault(require('lodash/union'));
var jsep = _interopDefault(require('jsep'));
var rollup = require('rollup');
var inject = _interopDefault(require('@mrgalaxy/rollup-plugin-inject'));

var babelHelpers = {};
babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

babelHelpers.classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

babelHelpers.createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

babelHelpers.defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

babelHelpers.inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

babelHelpers.possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

babelHelpers;

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * A cached reference to the hasOwnProperty function.
 */
var hasOwnProperty = Object.prototype.hasOwnProperty;


/**
 * A cached reference to the create function.
 */
var create = Object.create;


/**
 * Used to prevent property collisions between our "map" and its prototype.
 * @param {!Object<string, *>} map The map to check.
 * @param {string} property The property to check.
 * @return {boolean} Whether map has property.
 */
var has = function(map, property) {
  return hasOwnProperty.call(map, property);
};


/**
 * Creates an map object without a prototype.
 * @return {!Object}
 */
var createMap = function() {
  return create(null);
};

/**
 * Keeps track of information needed to perform diffs for a given DOM node.
 * @param {!string} nodeName
 * @param {?string=} key
 * @constructor
 */
function NodeData(nodeName, key) {
  /**
   * The attributes and their values.
   * @const {!Object<string, *>}
   */
  this.attrs = createMap();

  /**
   * An array of attribute name/value pairs, used for quickly diffing the
   * incomming attributes to see if the DOM node's attributes need to be
   * updated.
   * @const {Array<*>}
   */
  this.attrsArr = [];

  /**
   * The incoming attributes for this Node, before they are updated.
   * @const {!Object<string, *>}
   */
  this.newAttrs = createMap();

  /**
   * The key used to identify this node, used to preserve DOM nodes when they
   * move within their parent.
   * @const
   */
  this.key = key;

  /**
   * Keeps track of children within this node by their key.
   * {?Object<string, !Element>}
   */
  this.keyMap = null;

  /**
   * Whether or not the keyMap is currently valid.
   * {boolean}
   */
  this.keyMapValid = true;

  /**
   * The node name for this node.
   * @const {string}
   */
  this.nodeName = nodeName;

  /**
   * @type {?string}
   */
  this.text = null;
}


/**
 * Initializes a NodeData object for a Node.
 *
 * @param {Node} node The node to initialize data for.
 * @param {string} nodeName The node name of node.
 * @param {?string=} key The key that identifies the node.
 * @return {!NodeData} The newly initialized data object
 */
var initData = function(node, nodeName, key) {
  var data = new NodeData(nodeName, key);
  node['__incrementalDOMData'] = data;
  return data;
};


/**
 * Retrieves the NodeData object for a Node, creating it if necessary.
 *
 * @param {Node} node The node to retrieve the data for.
 * @return {!NodeData} The NodeData for this Node.
 */
var getData = function(node) {
  var data = node['__incrementalDOMData'];

  if (!data) {
    var nodeName = node.nodeName.toLowerCase();
    var key = null;

    if (node instanceof Element) {
      key = node.getAttribute('key');
    }

    data = initData(node, nodeName, key);
  }

  return data;
};

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var symbols = {
  default: '__default',

  placeholder: '__placeholder'
};

/**
 * Applies an attribute or property to a given Element. If the value is null
 * or undefined, it is removed from the Element. Otherwise, the value is set
 * as an attribute.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {?(boolean|number|string)=} value The attribute's value.
 */
var applyAttr = function(el, name, value) {
  if (value == null) {
    el.removeAttribute(name);
  } else {
    el.setAttribute(name, value);
  }
};

/**
 * Applies a property to a given Element.
 * @param {!Element} el
 * @param {string} name The property's name.
 * @param {*} value The property's value.
 */
var applyProp = function(el, name, value) {
  el[name] = value;
};


/**
 * Applies a style to an Element. No vendor prefix expansion is done for
 * property names/values.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} style The style to set. Either a string of css or an object
 *     containing property-value pairs.
 */
var applyStyle = function(el, name, style) {
  if (typeof style === 'string') {
    el.style.cssText = style;
  } else {
    el.style.cssText = '';
    var elStyle = el.style;
    var obj = /** @type {!Object<string,string>} */(style);

    for (var prop in obj) {
      if (has(obj, prop)) {
        elStyle[prop] = obj[prop];
      }
    }
  }
};


/**
 * Updates a single attribute on an Element.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} value The attribute's value. If the value is an object or
 *     function it is set on the Element, otherwise, it is set as an HTML
 *     attribute.
 */
var applyAttributeTyped = function(el, name, value) {
  var type = typeof value;

  if (type === 'object' || type === 'function') {
    applyProp(el, name, value);
  } else {
    applyAttr(el, name, /** @type {?(boolean|number|string)} */(value));
  }
};


/**
 * Calls the appropriate attribute mutator for this attribute.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} value The attribute's value.
 */
var updateAttribute = function(el, name, value) {
  var data = getData(el);
  var attrs = data.attrs;

  if (attrs[name] === value) {
    return;
  }

  var mutator = attributes[name] || attributes[symbols.default];
  mutator(el, name, value);

  attrs[name] = value;
};


/**
 * A publicly mutable object to provide custom mutators for attributes.
 * @const {!Object<string, function(!Element, string, *)>}
 */
var attributes = createMap();

// Special generic mutator that's called for any attribute that does not
// have a specific mutator.
attributes[symbols.default] = applyAttributeTyped;

attributes[symbols.placeholder] = function() {};

attributes['style'] = applyStyle;

/**
 * Gets the namespace to create an element (of a given tag) in.
 * @param {string} tag The tag to get the namespace for.
 * @param {?Node} parent
 * @return {?string} The namespace to create the tag in.
 */
var getNamespaceForTag = function(tag, parent) {
  if (tag === 'svg') {
    return 'http://www.w3.org/2000/svg';
  }

  if (getData(parent).nodeName === 'foreignObject') {
    return null;
  }

  return parent.namespaceURI;
};


/**
 * Creates an Element.
 * @param {Document} doc The document with which to create the Element.
 * @param {?Node} parent
 * @param {string} tag The tag for the Element.
 * @param {?string=} key A key to identify the Element.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element.
 * @return {!Element}
 */
var createElement = function(doc, parent, tag, key, statics) {
  var namespace = getNamespaceForTag(tag, parent);
  var el;

  if (namespace) {
    el = doc.createElementNS(namespace, tag);
  } else {
    el = doc.createElement(tag);
  }

  initData(el, tag, key);

  if (statics) {
    for (var i = 0; i < statics.length; i += 2) {
      updateAttribute(el, /** @type {!string}*/(statics[i]), statics[i + 1]);
    }
  }

  return el;
};


/**
 * Creates a Text Node.
 * @param {Document} doc The document with which to create the Element.
 * @return {!Text}
 */
var createText = function(doc) {
  var node = doc.createTextNode('');
  initData(node, '#text', null);
  return node;
};


/**
 * Creates a mapping that can be used to look up children using a key.
 * @param {?Node} el
 * @return {!Object<string, !Element>} A mapping of keys to the children of the
 *     Element.
 */
var createKeyMap = function(el) {
  var map = createMap();
  var children = el.children;
  var count = children.length;

  for (var i = 0; i < count; i += 1) {
    var child = children[i];
    var key = getData(child).key;

    if (key) {
      map[key] = child;
    }
  }

  return map;
};


/**
 * Retrieves the mapping of key to child node for a given Element, creating it
 * if necessary.
 * @param {?Node} el
 * @return {!Object<string, !Node>} A mapping of keys to child Elements
 */
var getKeyMap = function(el) {
  var data = getData(el);

  if (!data.keyMap) {
    data.keyMap = createKeyMap(el);
  }

  return data.keyMap;
};


/**
 * Retrieves a child from the parent with the given key.
 * @param {?Node} parent
 * @param {?string=} key
 * @return {?Node} The child corresponding to the key.
 */
var getChild = function(parent, key) {
  return key ? getKeyMap(parent)[key] : null;
};


/**
 * Registers an element as being a child. The parent will keep track of the
 * child using the key. The child can be retrieved using the same key using
 * getKeyMap. The provided key should be unique within the parent Element.
 * @param {?Node} parent The parent of child.
 * @param {string} key A key to identify the child with.
 * @param {!Node} child The child to register.
 */
var registerChild = function(parent, key, child) {
  getKeyMap(parent)[key] = child;
};

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** */
var notifications = {
  /**
   * Called after patch has compleated with any Nodes that have been created
   * and added to the DOM.
   * @type {?function(Array<!Node>)}
   */
  nodesCreated: null,

  /**
   * Called after patch has compleated with any Nodes that have been removed
   * from the DOM.
   * Note it's an applications responsibility to handle any childNodes.
   * @type {?function(Array<!Node>)}
   */
  nodesDeleted: null
};

/**
 * Keeps track of the state of a patch.
 * @constructor
 */
function Context() {
  /**
   * @type {(Array<!Node>|undefined)}
   */
  this.created = notifications.nodesCreated && [];

  /**
   * @type {(Array<!Node>|undefined)}
   */
  this.deleted = notifications.nodesDeleted && [];
}


/**
 * @param {!Node} node
 */
Context.prototype.markCreated = function(node) {
  if (this.created) {
    this.created.push(node);
  }
};


/**
 * @param {!Node} node
 */
Context.prototype.markDeleted = function(node) {
  if (this.deleted) {
    this.deleted.push(node);
  }
};


/**
 * Notifies about nodes that were created during the patch opearation.
 */
Context.prototype.notifyChanges = function() {
  if (this.created && this.created.length > 0) {
    notifications.nodesCreated(this.created);
  }

  if (this.deleted && this.deleted.length > 0) {
    notifications.nodesDeleted(this.deleted);
  }
};

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
  * Keeps track whether or not we are in an attributes declaration (after
  * elementOpenStart, but before elementOpenEnd).
  * @type {boolean}
  */
var inAttributes = false;


/**
  * Keeps track whether or not we are in an element that should not have its
  * children cleared.
  * @type {boolean}
  */
var inSkip = false;


/**
 * Makes sure that there is a current patch context.
 * @param {*} context
 */
var assertInPatch = function(context) {
  if (!context) {
    throw new Error('Cannot call currentElement() unless in patch');
  }
};


/**
* Makes sure that keyed Element matches the tag name provided.
* @param {!string} nodeName The nodeName of the node that is being matched.
* @param {string=} tag The tag name of the Element.
* @param {?string=} key The key of the Element.
*/
var assertKeyedTagMatches = function(nodeName, tag, key) {
  if (nodeName !== tag) {
    throw new Error('Was expecting node with key "' + key + '" to be a ' +
        tag + ', not a ' + nodeName + '.');
  }
};


/**
 * Makes sure that a patch closes every node that it opened.
 * @param {?Node} openElement
 * @param {!Node|!DocumentFragment} root
 */
var assertNoUnclosedTags = function(openElement, root) {
  if (openElement === root) {
    return;
  }

  var currentElement = openElement;
  var openTags = [];
  while (currentElement && currentElement !== root) {
    openTags.push(currentElement.nodeName.toLowerCase());
    currentElement = currentElement.parentNode;
  }

  throw new Error('One or more tags were not closed:\n' +
      openTags.join('\n'));
};


/**
 * Makes sure that the caller is not where attributes are expected.
 * @param {string} functionName
 */
var assertNotInAttributes = function(functionName) {
  if (inAttributes) {
    throw new Error(functionName + '() may not be called between ' +
        'elementOpenStart() and elementOpenEnd().');
  }
};


/**
 * Makes sure that the caller is not inside an element that has declared skip.
 * @param {string} functionName
 */
var assertNotInSkip = function(functionName) {
  if (inSkip) {
    throw new Error(functionName + '() may not be called inside an element ' +
        'that has called skip().');
  }
};


/**
 * Makes sure that the caller is where attributes are expected.
 * @param {string} functionName
 */
var assertInAttributes = function(functionName) {
  if (!inAttributes) {
    throw new Error(functionName + '() must be called after ' +
        'elementOpenStart().');
  }
};


/**
 * Makes sure the patch closes virtual attributes call
 */
var assertVirtualAttributesClosed = function() {
  if (inAttributes) {
    throw new Error('elementOpenEnd() must be called after calling ' +
        'elementOpenStart().');
  }
};


/**
  * Makes sure that placeholders have a key specified. Otherwise, conditional
  * placeholders and conditional elements next to placeholders will cause
  * placeholder elements to be re-used as non-placeholders and vice versa.
  * @param {string} key
  */
var assertPlaceholderKeySpecified = function(key) {
  if (!key) {
    throw new Error('Placeholder elements must have a key specified.');
  }
};


/**
  * Makes sure that tags are correctly nested.
  * @param {string} nodeName
  * @param {string} tag
  */
var assertCloseMatchesOpenTag = function(nodeName, tag) {
  if (nodeName !== tag) {
    throw new Error('Received a call to close ' + tag + ' but ' +
        nodeName + ' was open.');
  }
};


/**
 * Makes sure that no children elements have been declared yet in the current
 * element.
 * @param {string} functionName
 * @param {?Node} previousNode
 */
var assertNoChildrenDeclaredYet = function(functionName, previousNode) {
  if (previousNode !== null) {
    throw new Error(functionName + '() must come before any child ' +
        'declarations inside the current element.');
  }
};


/**
 * Updates the state of being in an attribute declaration.
 * @param {boolean} value
 * @return {boolean} the previous value.
 */
var setInAttributes = function(value) {
  var previous = inAttributes;
  inAttributes = value;
  return previous;
};


/**
 * Updates the state of being in a skip element.
 * @param {boolean} value
 * @return {boolean} the previous value.
 */
var setInSkip = function(value) {
  var previous = inSkip;
  inSkip = value;
  return previous;
};

/** @type {?Context} */
var context = null;

/** @type {?Node} */
var currentNode;

/** @type {?Node} */
var currentParent;

/** @type {?Node} */
var previousNode;

/** @type {?Element|?DocumentFragment} */
var root;

/** @type {?Document} */
var doc;


/**
 * Patches the document starting at el with the provided function. This function
 * may be called during an existing patch operation.
 * @param {!Element|!DocumentFragment} node The Element or Document
 *     to patch.
 * @param {!function(T)} fn A function containing elementOpen/elementClose/etc.
 *     calls that describe the DOM.
 * @param {T=} data An argument passed to fn to represent DOM state.
 * @template T
 */
var patch = function(node, fn, data) {
  var prevContext = context;
  var prevRoot = root;
  var prevDoc = doc;
  var prevCurrentNode = currentNode;
  var prevCurrentParent = currentParent;
  var prevPreviousNode = previousNode;
  var previousInAttributes = false;
  var previousInSkip = false;

  context = new Context();
  root = node;
  doc = node.ownerDocument;
  currentNode = node;
  currentParent = null;
  previousNode = null;

  if (process.env.NODE_ENV !== 'production') {
    previousInAttributes = setInAttributes(false);
    previousInSkip = setInSkip(false);
  }

  enterNode();
  fn(data);
  exitNode();

  if (process.env.NODE_ENV !== 'production') {
    assertVirtualAttributesClosed();
    assertNoUnclosedTags(previousNode, node);
    setInAttributes(previousInAttributes);
    setInSkip(previousInSkip);
  }

  context.notifyChanges();

  context = prevContext;
  root = prevRoot;
  doc = prevDoc;
  currentNode = prevCurrentNode;
  currentParent = prevCurrentParent;
  previousNode = prevPreviousNode;
};


/**
 * Checks whether or not the current node matches the specified nodeName and
 * key.
 *
 * @param {?string} nodeName The nodeName for this node.
 * @param {?string=} key An optional key that identifies a node.
 * @return {boolean} True if the node matches, false otherwise.
 */
var matches = function(nodeName, key) {
  var data = getData(currentNode);

  // Key check is done using double equals as we want to treat a null key the
  // same as undefined. This should be okay as the only values allowed are
  // strings, null and undefined so the == semantics are not too weird.
  return nodeName === data.nodeName && key == data.key;
};


/**
 * Aligns the virtual Element definition with the actual DOM, moving the
 * corresponding DOM node to the correct location or creating it if necessary.
 * @param {string} nodeName For an Element, this should be a valid tag string.
 *     For a Text, this should be #text.
 * @param {?string=} key The key used to identify this element.
 * @param {?Array<*>=} statics For an Element, this should be an array of
 *     name-value pairs.
 */
var alignWithDOM = function(nodeName, key, statics) {
  if (currentNode && matches(nodeName, key)) {
    return;
  }

  var node;

  // Check to see if the node has moved within the parent.
  if (key) {
    node = getChild(currentParent, key);
    if (node && process.env.NODE_ENV !== 'production') {
      assertKeyedTagMatches(getData(node).nodeName, nodeName, key);
    }
  }

  // Create the node if it doesn't exist.
  if (!node) {
    if (nodeName === '#text') {
      node = createText(doc);
    } else {
      node = createElement(doc, currentParent, nodeName, key, statics);
    }

    if (key) {
      registerChild(currentParent, key, node);
    }

    context.markCreated(node);
  }

  // If the node has a key, remove it from the DOM to prevent a large number
  // of re-orders in the case that it moved far or was completely removed.
  // Since we hold on to a reference through the keyMap, we can always add it
  // back.
  if (currentNode && getData(currentNode).key) {
    currentParent.replaceChild(node, currentNode);
    getData(currentParent).keyMapValid = false;
  } else {
    currentParent.insertBefore(node, currentNode);
  }

  currentNode = node;
};


/**
 * Clears out any unvisited Nodes, as the corresponding virtual element
 * functions were never called for them.
 */
var clearUnvisitedDOM = function() {
  var node = currentParent;
  var data = getData(node);
  var keyMap = data.keyMap;
  var keyMapValid = data.keyMapValid;
  var child = node.lastChild;
  var key;

  if (child === previousNode && keyMapValid) {
    return;
  }

  if (data.attrs[symbols.placeholder] && node !== root) {
    return;
  }

  while (child !== previousNode) {
    node.removeChild(child);
    context.markDeleted(/** @type {!Node}*/(child));

    key = getData(child).key;
    if (key) {
      delete keyMap[key];
    }
    child = node.lastChild;
  }

  // Clean the keyMap, removing any unusued keys.
  if (!keyMapValid) {
    for (key in keyMap) {
      child = keyMap[key];
      if (child.parentNode !== node) {
        context.markDeleted(child);
        delete keyMap[key];
      }
    }

    data.keyMapValid = true;
  }
};


/**
 * Changes to the first child of the current node.
 */
var enterNode = function() {
  currentParent = currentNode;
  currentNode = currentNode.firstChild;
  previousNode = null;
};


/**
 * Changes to the next sibling of the current node.
 */
var nextNode = function() {
  previousNode = currentNode;
  currentNode = currentNode.nextSibling;
};


/**
 * Changes to the parent of the current node, removing any unvisited children.
 */
var exitNode = function() {
  clearUnvisitedDOM();

  previousNode = currentParent;
  currentNode = currentParent.nextSibling;
  currentParent = currentParent.parentNode;
};


/**
 * Makes sure that the current node is an Element with a matching tagName and
 * key.
 *
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @return {!Element} The corresponding Element.
 */
var coreElementOpen = function(tag, key, statics) {
  alignWithDOM(tag, key, statics);
  enterNode();
  return /** @type {!Element} */(currentParent);
};


/**
 * Closes the currently open Element, removing any unvisited children if
 * necessary.
 *
 * @return {!Element} The corresponding Element.
 */
var coreElementClose = function() {
  if (process.env.NODE_ENV !== 'production') {
    setInSkip(false);
  }

  exitNode();
  return /** @type {!Element} */(previousNode);
};


/**
 * Makes sure the current node is a Text node and creates a Text node if it is
 * not.
 *
 * @return {!Text} The corresponding Text Node.
 */
var coreText = function() {
  alignWithDOM('#text', null, null);
  nextNode();
  return /** @type {!Text} */(previousNode);
};


/**
 * Gets the current Element being patched.
 * @return {!Element}
 */
var currentElement = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertInPatch(context);
    assertNotInAttributes('currentElement');
  }
  return /** @type {!Element} */(currentParent);
};


/**
 * Skips the children in a subtree, allowing an Element to be closed without
 * clearing out the children.
 */
var skip = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertNoChildrenDeclaredYet('skip', previousNode);
    setInSkip(true);
  }
  previousNode = currentParent.lastChild;
};

/**
 * The offset in the virtual element declaration where the attributes are
 * specified.
 * @const
 */
var ATTRIBUTES_OFFSET = 3;


/**
 * Builds an array of arguments for use with elementOpenStart, attr and
 * elementOpenEnd.
 * @const {Array<*>}
 */
var argsBuilder = [];


/**
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementOpen = function(tag, key, statics, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementOpen');
    assertNotInSkip('elementOpen');
  }

  var node = coreElementOpen(tag, key, statics);
  var data = getData(node);

  /*
   * Checks to see if one or more attributes have changed for a given Element.
   * When no attributes have changed, this is much faster than checking each
   * individual argument. When attributes have changed, the overhead of this is
   * minimal.
   */
  var attrsArr = data.attrsArr;
  var newAttrs = data.newAttrs;
  var attrsChanged = false;
  var i = ATTRIBUTES_OFFSET;
  var j = 0;

  for (; i < arguments.length; i += 1, j += 1) {
    if (attrsArr[j] !== arguments[i]) {
      attrsChanged = true;
      break;
    }
  }

  for (; i < arguments.length; i += 1, j += 1) {
    attrsArr[j] = arguments[i];
  }

  if (j < attrsArr.length) {
    attrsChanged = true;
    attrsArr.length = j;
  }

  /*
   * Actually perform the attribute update.
   */
  if (attrsChanged) {
    for (i = ATTRIBUTES_OFFSET; i < arguments.length; i += 2) {
      newAttrs[arguments[i]] = arguments[i + 1];
    }

    for (var attr in newAttrs) {
      updateAttribute(node, attr, newAttrs[attr]);
      newAttrs[attr] = undefined;
    }
  }

  return node;
};


/**
 * Declares a virtual Element at the current location in the document. This
 * corresponds to an opening tag and a elementClose tag is required. This is
 * like elementOpen, but the attributes are defined using the attr function
 * rather than being passed as arguments. Must be folllowed by 0 or more calls
 * to attr, then a call to elementOpenEnd.
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 */
var elementOpenStart = function(tag, key, statics) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementOpenStart');
    setInAttributes(true);
  }

  argsBuilder[0] = tag;
  argsBuilder[1] = key;
  argsBuilder[2] = statics;
};


/***
 * Defines a virtual attribute at this point of the DOM. This is only valid
 * when called between elementOpenStart and elementOpenEnd.
 *
 * @param {string} name
 * @param {*} value
 */
var attr = function(name, value) {
  if (process.env.NODE_ENV !== 'production') {
    assertInAttributes('attr');
  }

  argsBuilder.push(name, value);
};


/**
 * Closes an open tag started with elementOpenStart.
 * @return {!Element} The corresponding Element.
 */
var elementOpenEnd = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertInAttributes('elementOpenEnd');
    setInAttributes(false);
  }

  var node = elementOpen.apply(null, argsBuilder);
  argsBuilder.length = 0;
  return node;
};


/**
 * Closes an open virtual Element.
 *
 * @param {string} tag The element's tag.
 * @return {!Element} The corresponding Element.
 */
var elementClose = function(tag) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementClose');
  }

  var node = coreElementClose();

  if (process.env.NODE_ENV !== 'production') {
    assertCloseMatchesOpenTag(getData(node).nodeName, tag);
  }

  return node;
};


/**
 * Declares a virtual Element at the current location in the document that has
 * no children.
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementVoid = function(tag, key, statics, var_args) {
  var node = elementOpen.apply(null, arguments);
  elementClose.apply(null, arguments);
  return node;
};


/**
 * Declares a virtual Element at the current location in the document that is a
 * placeholder element. Children of this Element can be manually managed and
 * will not be cleared by the library.
 *
 * A key must be specified to make sure that this node is correctly preserved
 * across all conditionals.
 *
 * @param {string} tag The element's tag.
 * @param {string} key The key used to identify this element.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementPlaceholder = function(tag, key, statics, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertPlaceholderKeySpecified(key);
  }

  elementOpen.apply(null, arguments);
  skip();
  return elementClose.apply(null, arguments);
};


/**
 * Declares a virtual Text at this point in the document.
 *
 * @param {string|number|boolean} value The value of the Text.
 * @param {...(function((string|number|boolean)):string)} var_args
 *     Functions to format the value which are called only when the value has
 *     changed.
 * @return {!Text} The corresponding text node.
 */
var text = function(value, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('text');
    assertNotInSkip('text');
  }

  var node = coreText();
  var data = getData(node);

  if (data.text !== value) {
    data.text = /** @type {string} */(value);

    var formatted = value;
    for (var i = 1; i < arguments.length; i += 1) {
      formatted = arguments[i](formatted);
    }

    node.data = formatted;
  }

  return node;
};

function forEach(val, fn, scope) {
	if (val && typeof val.forEach === "function") val.forEach(fn, scope);else _forEach(val, fn, scope);
}

function map(val, fn, scope) {
	var res = [];
	forEach(val, function (v, i, l) {
		res.push(fn.call(scope, v, i, l));
	});
	return res;
}

// like jQuery's empty(), removes all children
function emptyNode(node) {
	while (node.lastChild) {
		node.removeChild(node.lastChild);
	}return node;
}

var entity = /&(?:#x[a-f0-9]+|#[0-9]+|[a-z0-9]+);?/ig;

function containsEntities(str) {
	return entity.test(str);
}

// converts html entities to unicode
var decodeEntities = function () {
	if (typeof document === "undefined") return;

	// this prevents any overhead from creating the object each time
	var element = document.createElement('div');

	return function decodeHTMLEntities(str) {
		str = str.replace(entity, function (m) {
			element.innerHTML = m;
			return element.textContent;
		});

		emptyNode(element);

		return str;
	};
}();

// convert html into DOM nodes
function toString(value) {
	if (typeof value !== "string") {
		value = value != null ? value.toString() : "";
	}

	return value;
}

function joinValues() {
	return _toArray(arguments).map(toString).join("");
}

function matches$1(node, selector) {
	if (_isArray(selector)) return selector.some(function (s) {
		return matches$1(node, s);
	});

	if (selector instanceof window.Node) {
		return node === selector;
	}

	if (typeof selector === "function") {
		return !!selector(node);
	}

	if (node.nodeType === window.Node.ELEMENT_NODE) {
		return matchesSelector(node, selector);
	}

	return false;
}

// export function getPropertyFromClass(obj, prop) {
// 	var val;
// 	let proto = Object.getPrototypeOf(obj);
//
// 	if (typeof prop === "string" && prop) {
// 		let p = prop;
// 		prop = c => c[p];
// 	}
//
// 	if (typeof prop !== "function") {
// 		throw new Error("Expecting function or string for property.");
// 	}
//
// 	while (proto) {
// 		val = merge.defaults(val, prop(proto.constructor));
// 		proto = Object.getPrototypeOf(proto);
// 	}
//
// 	return val;
// }

var utils = Object.freeze({
	forEach: forEach,
	map: map,
	emptyNode: emptyNode,
	containsEntities: containsEntities,
	decodeEntities: decodeEntities,
	toString: toString,
	joinValues: joinValues,
	matches: matches$1
});

function autotext(fn, that) {
	var node = coreText();

	function renderText(c) {
		c.node = node;
		var data = getData(node);
		var value = fn.call(that, node);

		if (data.text !== value) {
			data.text = value;
			node.data = toString(value);
		}
	}

	return Trackr.autorun(renderText);
}

function autoelement(node, fn) {
	return Trackr.autorun(function (c) {
		c.element = node;

		// catch errors that idom likes to throw
		var cur = void 0;
		try {
			cur = currentElement();
		} catch (e) {
			e;
		}

		// call straight or patch
		if (cur === node) fn(node, c);else patch(node, function () {
			return fn(node, c);
		});
	});
}

var execDeletedCallbacks = function execDeletedCallbacks(node) {
	var data = getData(node);
	var cbs = data.deletedCallbacks;
	if (!cbs) return;
	cbs.splice(0, cbs.length).forEach(function (fn) {
		fn.call(null, node);
	});
};

notifications.nodesDeleted = function (nodes) {
	_forEach(nodes, function (node) {
		if (node.childNodes) notifications.nodesDeleted(node.childNodes);
		execDeletedCallbacks(node);
	});
};

function onDestroy(node, fn) {
	if (typeof fn !== "function") {
		throw new Error("Expecting function.");
	}

	var data = getData(node);
	if (!data.deletedCallbacks) data.deletedCallbacks = [];
	data.deletedCallbacks.push(fn);
}

function removeDestroyListener(node, fn) {
	var data = getData(node);
	if (data.deletedCallbacks) {
		data.deletedCallbacks = _without(data.deletedCallbacks, fn);
	}
}

// function renderHTML(node) {
// 	if (node.type == "text") {
// 		return idom.text(node.content);
// 	}
//
// 	if (node.type == "tag") {
// 		var argsArray = [node.name, null, null];
//
// 		// convert attrs into a flat array
// 		for (var attr in node.attrs) {
// 			argsArray.push(attr);
// 			argsArray.push(node.attrs[attr]);
// 		}
//
// 		if (node.voidElement) {
// 			return idom.elementVoid.apply(idom, argsArray);
// 		} else {
// 			idom.elementOpen.apply(idom, argsArray);
//
// 			for (var i = 0; i < node.children.length; i++) {
// 				renderHTML(node.children[i]);
// 			}
//
// 			return idom.elementClose(node.name);
// 		}
// 	}
// }

// export function html(src) {
// 	var tree = parseHTML("<div>" + src + "</div>");
// 	if (_.isArray(tree)) tree = tree[0];
// 	return tree.children.map(renderHTML);
// }

function updateProperty(el, name, value) {
	var data = getData(el);
	var attrs = data.attrs;

	if (attrs[name] === value) {
		return;
	}

	el[name] = value;
	attrs[name] = value;
}

var idom = Object.freeze({
	updateAttribute: updateAttribute,
	getData: getData,
	autotext: autotext,
	autoelement: autoelement,
	onDestroy: onDestroy,
	removeDestroyListener: removeDestroyListener,
	updateProperty: updateProperty,
	patch: patch,
	currentElement: currentElement,
	skip: skip,
	elementVoid: elementVoid,
	elementOpenStart: elementOpenStart,
	elementOpenEnd: elementOpenEnd,
	elementOpen: elementOpen,
	elementClose: elementClose,
	elementPlaceholder: elementPlaceholder,
	text: text,
	attr: attr,
	symbols: symbols,
	attributes: attributes,
	applyAttr: applyAttr,
	applyProp: applyProp,
	notifications: notifications
});

var decorators = {};

function plugin() {
	this._decorators = {};
	this.decorate = add;
	this.stopDecorating = remove;
	this.findDecorator = find;
}

// creates a decorator
function add(name, fn, options) {
	var _this = this;

	if ((typeof name === "undefined" ? "undefined" : babelHelpers.typeof(name)) === "object") {
		options = fn;
		_forEach(name, function (fn, n) {
			if (_isArray(fn)) add.call(_this, n, fn[0], fn[1]);else add.call(_this, n, fn, options);
		});
		return this;
	}

	if (typeof name !== "string" || name === "") throw new Error("Expecting non-empty string for decorator name.");
	if (typeof fn !== "function") throw new Error("Expecting function for decorator.");

	var decs = this._decorators ? this._decorators : decorators;
	decs[name] = {
		template: this,
		callback: fn,
		options: options || {}
	};

	return this;
}

// removes a decorator
function remove(name) {
	if (this._decorators) {
		if (name == null) this._decorators = {};else delete this._decorators[name];
	} else if (name) {
		delete decorators[name];
	}

	return this;
}

function find(name) {
	var decs = this._decorators;
	if (decs != null && decs[name]) {
		return decs[name];
	}
}

// finds the best decorator
function lookup$1(scope, name) {
	// look on the closest template
	var owner = scope.getTemplateScope();
	if (owner) {
		var dec = owner.template.findDecorator(name);
		if (dec) return _assign({ owner: owner }, dec);
	}

	// look globally
	if (decorators[name]) {
		return _assign({ owner: global }, decorators[name]);
	}
}

function setAttribute(node, name, values) {
	var v = !values || !values.length || values[0] == null ? "" : values[0];
	updateAttribute(node, name, v);
}

function render$1(_scope, name, value) {
	var node = currentElement();
	if (!node) throw new Error("Not currently patching.");

	// each decorator is given its own scope
	var scope = new Scope(null, _scope);

	var isStaticValue = typeof value !== "function";
	var getValue = function getValue() {
		var val = isStaticValue ? value : value(scope);
		if (!_isArray(val)) val = [val];
		return val;
	};

	// look up decorator by name
	var d = lookup$1(scope, name);

	// quick escape if static value and no decorator
	if (isStaticValue && !d) {
		setAttribute(node, name, getValue());
		return;
	}

	var anim = void 0,
	    comp = void 0;
	var cancel = false;
	var decscope = _assign({
		target: node,
		scope: scope
	}, d);

	var run = function run(c) {
		decscope.comp = c;
		var val = getValue();

		if (!d) {
			setAttribute(node, name, val);
			return;
		}

		// execute the callback
		d.callback.apply(d.template, [decscope].concat(val));
	};

	// defer computation if desired
	if (d && d.options && d.options.defer) {
		// clean up when current computation reruns
		var pcomp = Trackr.currentComputation;
		if (!pcomp) throw new Error("Can only render decorator in a computation.");
		pcomp.onInvalidate(function () {
			cancel = true;
			if (anim) raf.cancel(anim);
			if (comp) comp.stop();
		});

		// run decorator on the next frame
		anim = raf(function () {
			if (cancel) return;
			anim = null;
			comp = Trackr.autorun(run);
		});
	} else {
		Trackr.autorun(run);
	}
}

var decorators$1 = Object.freeze({
	plugin: plugin,
	default: plugin,
	add: add,
	remove: remove,
	find: find,
	lookup: lookup$1,
	render: render$1
});

var decorators$2 = {};
var actions = {};

// Action Class
var Action = function () {
	function Action(name, scope) {
		babelHelpers.classCallCheck(this, Action);

		this.name = name;
		this.scope = scope;
		this.bubbles = true;
	}

	babelHelpers.createClass(Action, [{
		key: "stopPropagation",
		value: function stopPropagation() {
			this.bubbles = false;
			return this;
		}
	}], [{
		key: "create",
		value: function create(name, scope) {
			var action = void 0;

			if (typeof name === "string") {
				action = new Action(name, scope);
			} else if (_isObject(name) && !(name instanceof Action)) {
				action = _assign(new Action(), action);
			} else {
				action = name;
			}

			if (!(action instanceof Action)) {
				throw new Error("Expecting action name, object or instance of Action.");
			}

			return action;
		}
	}]);
	return Action;
}();

// the plugin
function plugin$1() {
	this.use("decorators");
	this._actions = {};
	this.actions = this.addAction = add$1;
	this.removeAction = remove$1;
	this.decorate(decorators$2);
}

// standard dom events
defineEvent(['load', 'scroll', 'click', 'dblclick', 'mousedown', 'mouseup', 'mouseenter', 'mouseleave', 'keydown', 'keypress', 'keyup', 'blur', 'focus', 'change', 'input', 'submit', 'reset', 'drag', 'dragdrop', 'dragend', 'dragenter', 'dragexit', 'dragleave', 'dragover', 'dragstart', 'drop']);

function defineEvent(event) {
	if (_isArray(event)) return event.forEach(defineEvent);

	decorators$2["on-" + event] = function (decor, key) {
		var node = void 0,
		    args = void 0;
		var listener = function listener(e) {
			// create a new action object
			var action = new Action(key, decor.scope);
			action.original = e;
			action.target = action.node = node;
			action.owner = decor.owner;
			action.template = decor.template;

			// fire the action
			return fire(action, null, args);
		};

		node = decor.target;
		args = Array.prototype.slice.call(arguments, 2);

		node.addEventListener(event, listener);
		decor.comp.onInvalidate(function () {
			node.removeEventListener(event, listener);
		});
	};
}

// Msutache Instance Methods
function add$1(name, fn) {
	var _this = this;

	if ((typeof name === "undefined" ? "undefined" : babelHelpers.typeof(name)) === "object" && fn == null) {
		_forEach(name, function (fn, n) {
			return add$1.call(_this, n, fn);
		});
		return this;
	}

	if (typeof name !== "string" || name === "") throw new Error("Expecting non-empty string for action name.");
	if (typeof fn !== "function") throw new Error("Expecting function for action.");

	var obj = this._actions;
	if (!obj) obj = actions;
	if (obj[name] == null) obj[name] = [];
	if (! ~obj[name].indexOf(fn)) obj[name].push(fn);

	return this;
}

function remove$1(name, fn) {
	if (typeof name === "function" && fn == null) {
		fn = name;
		name = null;
	}

	var obj = this._actions;
	if (!obj) obj = actions;

	if (name == null && fn == null) {
		// clear actions, but never on the global
		if (this._actions != null) this._actions = {};
	} else if (fn == null) {
		delete obj[name];
	} else if (name == null) {
		_forEach(obj, function (d, n) {
			obj[n] = d.filter(function (f) {
				return f !== fn;
			});
		});
	} else if (obj[name] != null) {
		obj[name] = _without(obj[name], fn);
	}

	return this;
}

function fire(a, b, args) {
	var action = Action.create(a, b);
	if (!action.scope) {
		throw new Error("Action is missing a scope.");
	}

	var scope = action.scope;
	var name = action.name;
	args = [].concat(action, args);

	// runs function, unless propagation is stopped
	function run(fn) {
		if (fn.apply(scope, args) === false) {
			action.bubbles = false;
		}

		return !action.bubbles;
	}

	// bubble the action up through all the scopes
	while (action.bubbles && scope) {
		if (scope.template) {
			var acts = scope.template._actions;
			if (acts != null && Array.isArray(acts[name])) {
				acts[name].some(run);
			}
		}

		scope = scope.parent;
	}

	// bubble action to the global actions
	if (action.bubbles && Array.isArray(actions[name])) {
		actions[name].some(run);
	}

	return Boolean(action.bubbles);
}

var actions$1 = Object.freeze({
	Action: Action,
	plugin: plugin$1,
	default: plugin$1,
	defineEvent: defineEvent,
	add: add$1,
	remove: remove$1,
	fire: fire
});

var value_types = ["radio", "option"];
var selection_types = ["text", "search", "tel", "url", "password"];

var formBindings = {};

function plugin$2(options) {
	this.use("decorators");

	options = options || {};

	// add methods
	this._formBindings = {};
	this.twoway = this.addFormBinding = add$2;
	this.getFormBinding = get$1;
	this.removeFormBinding = remove$2;

	// add main binding decorator
	this.decorate("bind-to", bindTo.bind(this, options, false));
	this.decorate("lazybind-to", bindTo.bind(this, options, true));

	// add value decorator for radios and options
	this.decorate("value", valueOf, { parse: "string" });
}

function add$2(id, getter, onChange) {
	var _this = this;

	if (_isObject(id)) {
		_forEach(id, function (v, k) {
			return add$2.call(_this, k, v);
		});
		return this;
	}

	if (typeof id !== "string") throw new Error("Expecting a string for the form binding ID.");

	var bindings = this._formBindings || formBindings;
	if (bindings[id] != null) throw new Error("A form binding with id '" + id + "' already exists.");

	if (_isObject(getter) && onChange == null) {
		onChange = getter.change;
		getter = getter.get;
	}

	if (typeof getter !== "function") throw new Error("Expecting a function or object for the form binding getter.");
	if (typeof onChange !== "function") onChange = null;

	bindings[id] = {
		get: getter,
		change: onChange
	};

	return this;
}

function get$1(id) {
	if (typeof id !== "string") return;

	if (this._formBindings && this._formBindings[id]) {
		return this._formBindings[id];
	}

	return formBindings[id];
}

function remove$2(id) {
	if (this._formBindings) {
		if (id == null) this._formBindings = {};else delete this._formBindings[id];
	} else if (id) {
		delete formBindings[id];
	}

	return this;
}

var type_map = {
	"text": ["text", "color", "email", "password", "search", "tel", "url", "hidden"],
	"number": ["number", "range"],
	"date": ["date", "datetime", "datetime-local", "month", "time", "week"],
	"file": ["file"],
	"checkbox": ["checkbox"],
	"radio": ["radio"]
};

function getType(el) {
	switch (el.tagName.toLowerCase()) {
		case "input":
			for (var type in type_map) {
				if (_includes(type_map[type], el.type)) return type;
			}
			break;

		case "select":
			return "select";

		case "option":
			return "option";

		case "textarea":
			return "text";
	}
}

function getNodeValue(node, type) {
	if (type == null) type = getType(node);
	var val;

	switch (type) {
		case "number":
			val = node.valueAsNumber;
			break;
		case "text":
			val = node.value;
			break;

		case "checkbox":
			val = node.checked;
			break;

		case "date":
			val = node.valueAsDate;
			break;

		case "select":
			var opt = node.querySelector("option:checked");
			if (opt != null) val = opt.$bound_value;
			break;

		case "file":
			val = !node.multiple ? node.files[0] : _toArray(node.files);
			break;

		case "radio":
			val = node.$bound_value;
			break;
	}

	return val;
}

function setLiveValue(live, el, fn) {
	var active = document.activeElement === el;
	if (!live && active) return;
	var sel = selection_types.indexOf(el.type) > -1;
	var pos = sel ? [el.selectionStart, el.selectionEnd] : null;
	fn();
	if (sel && active) el.setSelectionRange.apply(el, pos);
}

function setNodeValue(el, val, type, live) {
	if (type == null) type = getType(el);

	switch (type) {
		case "number":
			setLiveValue(live, el, function () {
				updateProperty(el, typeof val === "number" ? "valueAsNumber" : "value", val);
			});
			break;

		case "text":
			setLiveValue(live, el, function () {
				updateProperty(el, "value", toString(val));
			});
			break;

		case "checkbox":
			updateProperty(el, "checked", Boolean(val));
			break;

		case "date":
			setLiveValue(live, el, function () {
				updateProperty(el, _isDate(val) ? "valueAsDate" : "value", val);
			});
			break;

		case "select":
			_toArray(el.querySelectorAll("option")).forEach(function (opt) {
				updateProperty(opt, "selected", opt.$bound_value === val);
			});
			break;

		case "radio":
			updateProperty(el, "checked", el.$bound_value === val);
			break;
	}
}

function bindTo(options, lazy, d, id) {
	var fbind = this.getFormBinding(id);
	if (fbind == null) return;

	var el = d.target;
	var args = _toArray(arguments).slice(4);
	var type = getType(el);
	var onChange = void 0;
	var twscope = _pick(d, "scope", "template", "target");

	// detect changes to the input's value
	if (typeof fbind.change === "function") {
		onChange = function onChange(e) {
			fbind.change.apply(_assign(twscope, {
				original: e
			}), [getNodeValue(el, type)].concat(args));
		};

		el.addEventListener("change", onChange);
		el.addEventListener("paste", onChange);
		if (!(options.lazy || lazy)) el.addEventListener("keyup", onChange);

		d.comp.onInvalidate(function () {
			el.removeEventListener("change", onChange);
			el.removeEventListener("paste", onChange);
			el.removeEventListener("keyup", onChange);
		});
	}

	var nodeValueComp = void 0,
	    stopped = false;

	d.comp.onInvalidate(function () {
		stopped = true;
		if (nodeValueComp) nodeValueComp.stop();
	});

	// reactively set the value on the input
	// deferred so value decorators run
	_defer(function () {
		if (stopped) return;
		nodeValueComp = Trackr.autorun(function () {
			setNodeValue(el, fbind.get.apply(twscope, args), type, options.live);
		});
	});
}

function valueOf(d, val) {
	var el = d.target,
	    type = getType(el);

	if (!_includes(value_types, type)) {
		updateAttribute(el, "value", val);
		return;
	}

	el.$bound_value = val;
	updateProperty(el, "value", val);
}

function plugin$3() {
	this.use("decorators");
	this.decorate("ref", decorator);
}

var warned = [];

function warnOnce(msg) {
	if (_includes(warned, msg)) return;
	console.warn(msg);
	warned.push(msg);
}

function decorator(d, key) {
	var scope = d.owner;
	if (!scope.refs) scope.refs = {};

	// warn about overwrites overwrite
	if (_has(scope.refs, key)) {
		warnOnce("Multiple elements with reference '" + key + "'.");
		return;
	}

	// set the reference
	scope.refs[key] = d.target;

	// remove the reference when the element disappears
	d.comp.onInvalidate(function () {
		delete scope.refs[key];
	});
}

// export function find(key) {
// 	var tpls = [ this ],
// 		tpl;
//
// 	while (tpls.length) {
// 		tpl = tpls.shift();
// 		if (tpl.refs && tpl.refs[key]) return tpl.refs[key];
// 		// tpls = tpls.concat(tpl.getComponents());
// 	}
//
// 	return null;
// }

// declare, but don't set the plugin variable
// exports and imports are hoisted, so this cannot do anything to the value
var plugins;

function load(tpl, plugin, args) {
	var name = void 0;
	if (plugins == null) plugins = {};

	if (_isString(plugin)) {
		name = plugin;
		if (plugins[name] == null) throw new Error("No plugin exists with id '" + name + "'.");

		plugin = plugins[name];
	}

	if (!_isFunction(plugin)) throw new Error("Expecting string or function for plugin");

	// check if plugin is already loaded on this template
	if (tpl._loaded_plugins == null) tpl._loaded_plugins = [];
	if (tpl._loaded_plugins.some(function (p) {
		return p.plugin === plugin;
	})) return tpl;

	if (args == null) args = [];
	if (!_isArray(args)) args = [args];
	tpl._loaded_plugins.push({
		name: name,
		plugin: plugin,
		args: args
	});

	plugin.apply(tpl, args);
	return tpl;
}

function register(name, fn) {
	if (plugins == null) plugins = {};

	if (typeof name !== "string") {
		throw new Error("Expecting string name for plugin.");
	}

	if (typeof fn !== "function") {
		throw new Error("Expecting function for plugin.");
	}

	if (fn === plugins[name]) return;
	if (plugins[name] != null) {
		throw new Error("Refusing to overwrite existing plugin \"name\".");
	}

	plugins[name] = fn;
}

function get(name) {
	return plugins && plugins[name];
}

register("decorators", plugin);
register("actions", plugin$1);
register("twoway", plugin$2);
register("refs", plugin$3);

var plugins$1 = Object.freeze({
	load: load,
	register: register,
	get: get
});

var templates = {};
var types = {};

function Template(name, render) {
	if (typeof name !== "string" || name === "") {
		throw new Error("Expecting a non-empty string for template name.");
	}

	if (typeof render !== "function") {
		throw new Error("Expecting a function for render.");
	}

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	// internal state
	this.s = {
		// template name
		name: name,
		// render method
		render: render,
		// dependency for easy invalidate
		renderdep: new Trackr.Dependency(),
		// holds the helpers
		helpers: new trackrObjects.Map()
	};

	// default plugins
	this.use("decorators");
	this.initialize();
}

Template.extend = subclass;
Template.prototype = Object.create(events.EventEmitter.prototype);
Template.prototype.constructor = Template;
Template.prototype.type = "template";
Template.prototype.initialize = function () {};

// plugin proxy for scopes
Template.prototype.use = function use(p) {
	return load(this, p, _toArray(arguments).slice(1));
};

Template.prototype.createScope = function (data, parent) {
	if (data instanceof Scope) {
		;
		var _ref = [data, null];
		parent = _ref[0];
		data = _ref[1];
	}var scope = new Scope(data, parent, this);
	this.emit("scope", scope);
	return scope;
};

Template.prototype.invalidate = function () {
	this.s.renderdep.changed();
	return this;
};

Template.prototype.render = function (data, key) {
	this.s.renderdep.depend();
	var scope = this.createScope(data);
	this.s.render(scope, key);
	this.emit("render", scope);
	return scope;
};

var injectable_nodes = [1, // Node.ELEMENT_NODE,
9, // Node.DOCUMENT_NODE,
11 // Node.DOCUMENT_FRAGMENT_NODE
];

Template.prototype.paint = function (node, data) {
	var _this = this;

	if (typeof node === "string") node = document.querySelector(node);
	if (!node || !_includes(injectable_nodes, node.nodeType)) {
		throw new Error("Expecting a valid DOM element to paint.");
	}

	var scope = data instanceof Scope ? data : new Scope(data);
	var c = Trackr.autorun(function () {
		patch(node, function () {
			return _this.render(scope);
		});
	});

	c.scope = scope;
	c.template = this;

	c.onStop(function () {
		patch(node, function () {});
	});

	return c;
};

var globalHelpers = new trackrObjects.Map();

var helpers = Template.prototype.helpers = function (key, value) {
	var _this2 = this;

	if ((typeof key === "undefined" ? "undefined" : babelHelpers.typeof(key)) === "object") {
		Object.keys(key).forEach(function (k) {
			return helpers.call(_this2, k, key[k]);
		});
		return this;
	}

	var h = globalHelpers;
	if (this instanceof Template) h = this.s.helpers;

	if (typeof value === "undefined") h.delete(key);else h.set(key, value);

	return this;
};

var getHelper = Template.prototype.getHelper = function (key) {
	if (this instanceof Template) {
		return this.s.helpers.get(key);
	}

	return globalHelpers.get(key);
};

function render(name, scope, key) {
	var tpl = getByName(name);
	if (tpl) return tpl.render(scope, key);
}

function getByName(name) {
	return templates[name];
}

function paint(tpl, node, data) {
	if (typeof tpl === "string") {
		tpl = getByName(tpl);
		if (!tpl) throw new Error("No template exists with name '" + tpl + "'");
	}
	if (!(tpl instanceof Template)) {
		throw new Error("Expecting template name or instance of Template.");
	}
	return tpl.paint(node, data);
}

function registerType(type, props) {
	if (typeof type !== "string" || !type) {
		throw new Error("Expecting non-empty string for type.");
	}
	if (_has(types, type)) {
		throw new Error("Template type '" + type + "' already exists.");
	}

	if (typeof props === "function") {
		props = { initialize: props };
	}

	props = _assign({}, props, { type: type });

	var T = Template;
	if (props.extends && _has(types, props.extends)) {
		T = types[props.extends];
		delete props.extends;
	}

	function wrap(k, f) {
		return function () {
			var osuper = this.super;
			this.super = T.prototype[k];
			var ret = f.apply(this, arguments);
			this.super = osuper;
			return ret;
		};
	}

	for (var k in props) {
		var fn = props[k];
		if (typeof fn !== "function") continue;
		if (/this\.super|this\["super"\]|this\['super'\]/.test(fn.toString())) {
			props[k] = wrap(k, fn);
		}
	}

	return types[type] = T.extend(props);
}

function create$1(name, type, render) {
	if (typeof name !== "string" || !name) {
		throw new Error("Expecting non-empty string for name.");
	}
	if (getByName(name)) {
		throw new Error("Template '" + name + "' already exists.");
	}

	if (typeof type === "function") {
		;
		var _ref2 = [type, null];
		render = _ref2[0];
		type = _ref2[1];
	}if (type && !_has(types, type)) {
		throw new Error("Template type " + type + " does not exist.");
	}

	var T = type ? types[type] : Template;
	return templates[name] = new T(name, render);
}

function Scope(data, parent, template) {
	if (!(this instanceof Scope)) {
		return new Scope(parent);
	}

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	if (data instanceof Scope) {
		var _ref = [parent, data, null];
		template = _ref[0];
		parent = _ref[1];
		data = _ref[2];
	}

	if (parent && !(parent instanceof Scope)) {
		throw new Error("Expecting an instance of Scope for parent.");
	}

	// internal view state
	this.s = {
		// parent view for scope
		parent: parent || null,
		// associated template object
		template: template || null,
		// holds lexical data
		scope: {},
		// holds "this" data
		data: new trackrObjects.Variable()
	};

	// set initial data
	if (data != null) this.set(data);
	if (!this.parent) this.dataVar.set(data);
}

Scope.prototype = Object.create(events.EventEmitter.prototype);
Scope.prototype.constructor = Scope;

assignProps(Scope.prototype, {
	scope: function scope() {
		return this.s.scope;
	},
	dataVar: function dataVar() {
		return this.s.data;
	},
	data: function data() {
		return this.s.data.get();
	},
	template: function template() {
		return this.s.template;
	},
	parent: function parent() {
		return this.s.parent;
	}
});

Scope.prototype.set = function (key, value) {
	var _this = this;

	if ((typeof key === "undefined" ? "undefined" : babelHelpers.typeof(key)) === "object") {
		_forEach(key, function (v, k) {
			return _this.set(k, v);
		});
		return this;
	}

	if (key === "this") {
		this.s.data.set(value);
	} else {
		if (typeof value === "undefined") delete this.s.scope[key];else this.s.scope[key] = value;
	}

	return this;
};

Scope.prototype.get = function (key) {
	return _has(this.s.scope, key) ? this.s.scope[key] : void 0;
};

Scope.prototype.parentData = function (dist) {
	if (typeof dist !== "number" || isNaN(dist)) dist = 1;

	if (dist >= 0) {
		var _v = this;
		while (dist && _v) {
			dist--;
			_v = _v.parent;
		}
		return !dist && _v ? _v.data : void 0;
	}

	var views = [];
	var v = this;

	while (v) {
		views.push(v);
		v = v.parent;
	}

	var view = views[Math.abs(dist) - 1];
	return view ? view.data : void 0;
};

Scope.prototype.getTemplateScope = function () {
	var view = this;

	while (view) {
		if (view.template != null) {
			return view;
		}

		view = view.parent;
	}
};

Scope.prototype.getTemplate = function () {
	var scope = this.getTemplateScope();
	return scope ? scope.template : void 0;
};

var lookup = function lookup(tpl, scope, key) {
	var view = void 0;

	// 0-0. check for this
	if (!key || key === "this") {
		view = scope;
		while (view) {
			var _val = view.data;
			if (_val !== void 0) return _val;
			view = view.parent;
		}
	}

	// 0-1. special method $this
	if (key === "$this") {
		return function (d) {
			return scope.parentData(d);
		};
	}

	// 1. check closest template helpers
	if (tpl) {
		var _val2 = tpl.getHelper(key);
		if (_val2 !== void 0) return _val2;
	}

	// 2. check lexical scope
	view = scope;
	while (view) {
		var _val3 = view.get(key);
		if (_val3 !== void 0) return _val3;
		view = view.parent;
	}

	// 3. check global helpers
	var val = getHelper(key);
	if (val !== void 0) return val;
};

Scope.prototype.lookup = function (key) {
	return lookup(this.getTemplate(), this, key);
};

var ASTNode = function () {
	function ASTNode(line, col, props) {
		babelHelpers.classCallCheck(this, ASTNode);

		if ((typeof line === "undefined" ? "undefined" : babelHelpers.typeof(line)) === "object" && col == null && props == null) {
			var _ref = [line, null];
			props = _ref[0];
			line = _ref[1];
		}

		this._line = typeof line === "number" ? line : 0;
		this._column = typeof col === "number" ? col : 0;
		_assign(this, props);
	}

	babelHelpers.createClass(ASTNode, [{
		key: "_sn",
		value: function _sn(file, chunk) {
			return new sourceMap.SourceNode(this._line, this._column, file, chunk);
		}
	}, {
		key: "start",
		value: function start(data) {
			this._writer = {
				chunks: [],
				data: data
			};

			return this;
		}
	}, {
		key: "_normalize_indent",
		value: function _normalize_indent() {
			var d = this._writer.data;
			if (typeof d.indent === "string") d.tabchar = d.indent;
			if (d.indent !== false && typeof d.indent !== "number") d.indent = 0;
		}
	}, {
		key: "indent",
		value: function indent() {
			this._normalize_indent();
			var d = this._writer.data;
			if (typeof d.indent === "number") d.indent++;
			return this;
		}
	}, {
		key: "outdent",
		value: function outdent() {
			this._normalize_indent();
			var d = this._writer.data;
			if (typeof d.indent === "number" && d.indent > 0) d.indent--;
			return this;
		}
	}, {
		key: "tabs",
		value: function tabs() {
			this._normalize_indent();

			var res = "";
			var d = this._writer.data;
			var tabchar = d.tabchar;
			if (tabchar == null) tabchar = "  ";

			if (typeof d.indent === "number") {
				for (var i = 0; i < d.indent; i++) {
					res += tabchar;
				}
			}

			return res;
		}
	}, {
		key: "write",
		value: function write(chunk) {
			this.push([].concat(this.tabs(), chunk, "\n"));
			return this;
		}
	}, {
		key: "push",
		value: function push(chunk) {
			this._writer.chunks.push(chunk);
			return this;
		}
	}, {
		key: "end",
		value: function end(chunk) {
			if (chunk) this.write(chunk);
			var w = this._writer;
			delete this._writer;
			return this._sn(w.data.filename, w.chunks);
		}
	}, {
		key: "compile",
		value: function compile() {
			throw new Error("Not implemented.");
			// return this._sn(data.filename);
		}
	}, {
		key: "toJSON",
		value: function toJSON() {
			var _this = this;

			var out = {};
			Object.keys(this).forEach(function (k) {
				if (k[0] === "_") return;
				out[k] = _this[k];
			});
			_assign(out, _pick(this, "_line", "_column"));
		}
	}]);
	return ASTNode;
}();

var File = function (_Node) {
	babelHelpers.inherits(File, _Node);

	function File() {
		babelHelpers.classCallCheck(this, File);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(File).apply(this, arguments));
	}

	babelHelpers.createClass(File, [{
		key: "compile",
		value: function compile(data) {
			data = _assign({}, data, {
				filename: this.filename || _uniqueId("template") + ".html",
				headers: []
			});

			this.start(data);
			this.push(_invokeMap(this.children, "compile", data));

			var source = this.end();
			if (this.source) source.setSourceContent(data.filename, this.source);

			return source;
		}
	}]);
	return File;
}(ASTNode);

var Include = function (_Node) {
	babelHelpers.inherits(Include, _Node);

	function Include() {
		babelHelpers.classCallCheck(this, Include);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Include).apply(this, arguments));
	}

	babelHelpers.createClass(Include, [{
		key: "resolve",
		value: function resolve(file) {
			return path.resolve(path.dirname(file), this.src);
		}
	}, {
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.write("import " + JSON.stringify(this.src) + ";");
			return this.end();
		}
	}]);
	return Include;
}(ASTNode);

function header(data, h) {
	if (_isArray(data.headers) && !_includes(data.headers, h)) {
		data.headers.push(h);
	}
}

function hash(str) {
	var h = 0,
	    i,
	    chr,
	    len;

	if (str.length === 0) return h;

	for (i = 0, len = str.length; i < len; i++) {
		chr = str.charCodeAt(i);
		h = (h << 5) - h + chr;
		h |= 0; // Convert to 32bit integer
	}

	return h;
}

function compileGroup(nodes, data) {
	return nodes.map(function (n, i) {
		return n.compile(addKey(data, i.toString()));
	});
}

function compileGroupAsync(nodes, data) {
	nodes = nodes.slice(0);
	var out = [];
	var next = function next() {
		return !nodes.length ? Promise.resolve() : nodes.shift().compile(data).then(function (o) {
			out.push(o);
			return next();
		});
	};
	return next().then(function () {
		return out;
	});
}

function getKey(data) {
	return (data.key || []).reduce(function (memo, v) {
		if (_isString(v) && _isString(_last(memo))) {
			memo.push(memo.pop() + "-" + v);
		} else {
			memo.push(v);
		}

		return memo;
	}, []).map(function (v, i, l) {
		if ((typeof v === "undefined" ? "undefined" : babelHelpers.typeof(v)) === "object") return v.value;
		if (i !== 0) v = "-" + v;
		if (i !== l.length - 1) v += "-";
		return JSON.stringify(v);
	}).join(" + ");
}

function addKey(data, key) {
	var base = data.key || [];
	return _assign({}, data, {
		key: [].concat(base, key)
	});
}

function resetKey(data) {
	return _assign({}, data, { key: [] });
}

var utils$1 = Object.freeze({
	header: header,
	hash: hash,
	compileGroup: compileGroup,
	compileGroupAsync: compileGroupAsync,
	getKey: getKey,
	addKey: addKey,
	resetKey: resetKey
});

var Template$1 = function (_Node) {
	babelHelpers.inherits(Template, _Node);

	function Template() {
		babelHelpers.classCallCheck(this, Template);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Template).apply(this, arguments));
	}

	babelHelpers.createClass(Template, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);
			var safename = JSON.stringify(this.name);
			var safetype = this.type ? JSON.stringify(this.type) : null;

			header(data, "var Template = {};\n");

			this.write("Template[" + safename + "] = Temple.create(" + safename + ", " + (safetype ? safetype + ", " : "") + "function(scope, key) {").indent();
			data = addKey(data, { value: "key" });
			this.push(compileGroup(this.children, data));
			this.outdent().write("});\n");

			if (this.plugins && this.plugins.length) {
				this.plugins.forEach(function (p) {
					_this2.write("Template[" + safename + "].use(" + JSON.stringify(p) + ");");
				});
				this.write("");
			}

			return this.end();
		}
	}]);
	return Template;
}(ASTNode);

var Render = function (_Node) {
	babelHelpers.inherits(Render, _Node);

	function Render() {
		babelHelpers.classCallCheck(this, Render);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Render).apply(this, arguments));
	}

	babelHelpers.createClass(Render, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			var exp = this.expression.compile(data);
			this.write(["Temple.render(", exp, ", scope, " + getKey(data) + ");"]);
			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Render;
}(ASTNode);

var Text = function (_Node) {
	babelHelpers.inherits(Text, _Node);

	function Text() {
		babelHelpers.classCallCheck(this, Text);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Text).apply(this, arguments));
	}

	babelHelpers.createClass(Text, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var value = this.value.replace(/\s+/g, " ");
			var safe = JSON.stringify(value);

			if (containsEntities(value)) {
				safe = "Temple.utils.decodeEntities(" + safe + ")";
			}

			header(data, "var idom = Temple.idom;\n");
			this.write("idom.text(" + safe + ");");
			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Text;
}(ASTNode);

var Element$1 = function (_Node) {
	babelHelpers.inherits(Element, _Node);

	function Element() {
		babelHelpers.classCallCheck(this, Element);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Element).apply(this, arguments));
	}

	babelHelpers.createClass(Element, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);

			var tagName = JSON.stringify(this.tagname);
			var key = getKey(data);
			var childs = this.children;
			var attrs = this.attributes;

			var body = function body() {
				_this2.push([].concat(_invokeMap(attrs, "compile", data), compileGroup(childs, resetKey(data))));
			};

			header(data, "var idom = Temple.idom;\n");

			if (childs.length || attrs.length) {
				if (!childs.some(function (c) {
					return c.reactive;
				})) {
					this.write("idom.elementOpen(" + tagName + (key ? ", " + key : "") + ");");
					body();
					this.write("idom.elementClose(" + tagName + ");");
				} else {
					this.write("Temple.Element(" + tagName + ", " + (key || "null") + ", scope, function(scope) {").indent();
					body();
					this.outdent().write("});");
				}
			} else {
				this.write("idom.elementVoid(" + tagName + (key ? ", " + key : "") + ");");
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Element;
}(ASTNode);

var Attribute = function (_Node) {
	babelHelpers.inherits(Attribute, _Node);

	function Attribute() {
		babelHelpers.classCallCheck(this, Attribute);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Attribute).apply(this, arguments));
	}

	babelHelpers.createClass(Attribute, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var fun = void 0;
			var value = this._sn(data.filename, "");
			var len = this.children.length;

			if (len) {
				value.add(_invokeMap(this.children, "compile", data)).join(",");
				fun = this.children.some(function (c) {
					return c.reactive;
				});
				if (len > 1) value.prepend("[").add("]");

				if (fun && this.type === "string") {
					if (len > 1) value.add(".map(Temple.utils.toString).join(\"\")");else value.prepend("Temple.utils.toString(").add(")");
				}
			} else {
				value = null;
			}

			header(data, "var decorators = Temple.decorators;\n");
			this.push([this.tabs(), "decorators.render(scope, " + JSON.stringify(this.name)]);

			if (fun) {
				this.push(", function(scope) {\n").indent();
				this.write(["return ", value, ";"]);
				this.outdent().push([this.tabs(), "}"]);
			} else if (value) {
				this.push([", ", value]);
			}

			this.push(");\n");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Attribute;
}(ASTNode);

var If = function (_Node) {
	babelHelpers.inherits(If, _Node);

	function If() {
		babelHelpers.classCallCheck(this, If);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(If).apply(this, arguments));
	}

	babelHelpers.createClass(If, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);

			if (this.attribute) {
				this.push("(");
			}

			this.children.forEach(function (c, i, l) {
				if (_this2.attribute) {
					if (c.expression) _this2.push([c.expression.compile(data), " ? "]);
					_this2.push(["Temple.utils.joinValues(", c.compile(addKey(data, i.toString())).join(","), ")"]);
					if (i + 1 === l.length && c.expression) _this2.push(" : ''");
				} else {
					var line = [];
					line.push(!i ? "if" : !c.expression ? "else" : "else if", " ");
					if (c.expression) line.push(["(", c.expression.compile(data), ") "]);
					line.push("{");
					_this2.write(line).indent();
					_this2.push(c.compile(addKey(data, i.toString())));
					_this2.outdent().write("}");
				}
			});

			if (this.attribute) {
				this.push(")");
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return If;
}(ASTNode);

var Branch = function (_Node) {
	babelHelpers.inherits(Branch, _Node);

	function Branch() {
		babelHelpers.classCallCheck(this, Branch);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Branch).apply(this, arguments));
	}

	babelHelpers.createClass(Branch, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.push(compileGroup(this.children, data));
			return this.end();
		}
	}]);
	return Branch;
}(ASTNode);

var Each = function (_Node) {
	babelHelpers.inherits(Each, _Node);

	function Each() {
		babelHelpers.classCallCheck(this, Each);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Each).apply(this, arguments));
	}

	babelHelpers.createClass(Each, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);
			var indexvar = _uniqueId("index");
			var vars = JSON.stringify(this.variables || []);

			if (!this.attribute) this.push(this.tabs());
			this.push(["Temple.Each(", exp, ", ", vars, ", scope, function(scope, " + indexvar + ") {"]).indent();
			if (!this.attribute) this.push("\n");
			data = addKey(data, { value: indexvar });
			var c = this._sn(data.filename, compileGroup(this.children, data));
			if (this.attribute) c = [" return Temple.utils.joinValues(", c.join(","), "); "];
			this.push(c);
			this.outdent();
			if (!this.attribute) this.write("});");else this.push("}).join(\"\")");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Each;
}(ASTNode);

var With = function (_Node) {
	babelHelpers.inherits(With, _Node);

	function With() {
		babelHelpers.classCallCheck(this, With);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(With).apply(this, arguments));
	}

	babelHelpers.createClass(With, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);

			if (!this.attribute) this.push(this.tabs());
			this.push(["Temple.With(", exp, ", scope, function(scope) {"]).indent();
			if (!this.attribute) this.push("\n");
			var c = this._sn(data.filename, compileGroup(this.children, data));
			if (this.attribute) c = [" return Temple.utils.joinValues(", c.join(","), "); "];
			this.push(c);
			this.outdent();

			if (!this.attribute) this.write("});");else this.push("})");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return With;
}(ASTNode);

var Interpolator = function (_Node) {
	babelHelpers.inherits(Interpolator, _Node);

	function Interpolator() {
		babelHelpers.classCallCheck(this, Interpolator);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Interpolator).apply(this, arguments));
	}

	babelHelpers.createClass(Interpolator, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			header(data, "var idom = Temple.idom;\n");
			this.write("idom.autotext(" + this.expression.compile(_assign({ asFn: true }, data)) + ");");
			return this.end();
		}
	}, {
		key: "reactive",

		// while technically reactive, they contain their own autorun statement
		get: function get() {
			return false;
		}
	}]);
	return Interpolator;
}(ASTNode);

var Comment = function (_Node) {
	babelHelpers.inherits(Comment, _Node);

	function Comment() {
		babelHelpers.classCallCheck(this, Comment);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Comment).apply(this, arguments));
	}

	babelHelpers.createClass(Comment, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			return this.end();
		}
	}]);
	return Comment;
}(ASTNode);

// method to run to query the scope
var SCOPE_LOOKUP = "scope.lookup";

var Expression = function (_Node) {
	babelHelpers.inherits(Expression, _Node);

	function Expression() {
		babelHelpers.classCallCheck(this, Expression);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Expression).apply(this, arguments));
	}

	babelHelpers.createClass(Expression, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			var exp = Expression.render(this.tree);

			if (data.asFn) {
				this.push("function() {\n").indent();
				this.write("return " + exp + ";");
				this.outdent().push([this.tabs(), "}"]);
			} else {
				this.push(exp);
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return !Expression.isStatic(this.tree);
		}
	}], [{
		key: "isStatic",
		value: function isStatic(tree) {
			var _isStatic = Expression.isStatic;

			switch (tree.type) {
				case "Compound":
					return tree.body.every(_isStatic);
				case "Identifier":
					return false;
				case "ThisExpression":
					return false;
				case "MemberExpression":
					{
						if (!_isStatic(tree.object)) return false;
						var prop = tree.property;
						if (!tree.computed && prop.type === "Identifier") return true;
						return _isStatic(prop);
					}
				case "Literal":
					return true;
				case "CallExpression":
					{
						return _isStatic(tree.callee) && tree.arguments.every(_isStatic);
					}
				case "UnaryExpression":
					return _isStatic(tree.argument);
				case "LogicalExpression":
				case "BinaryExpression":
					return _isStatic(tree.left) && _isStatic(tree.right);
				case "ArrayExpression":
					return tree.elements.every(_isStatic);
				case "ConditionalExpression":
					{
						return [tree.test, tree.consequent, tree.alternate].every(_isStatic);
					}
				default:
					return false;
			}
		}
	}, {
		key: "render",
		value: function render(tree) {
			var _render = Expression.render;

			switch (tree.type) {
				case "Compound":
					{
						// expressions seperated by comma or semicolon
						return "(" + tree.body.map(_render).join(",") + ")";
					}

				case "Identifier":
					// plain varable
					return SCOPE_LOOKUP + "(" + JSON.stringify(tree.name) + ")";

				case "ThisExpression":
					// plain varable
					return SCOPE_LOOKUP + "()";

				case "MemberExpression":
					{
						// object property
						var out = _render(tree.object);
						var prop = tree.property;
						if (!tree.computed && prop.type === "Identifier") out += "." + prop.name;else out += "[" + _render(prop) + "]";
						return out;
					}

				case "Literal":
					// string, number, boolean
					return JSON.stringify(tree.value);

				case "CallExpression":
					{
						// function call
						var _out = _render(tree.callee);
						if (tree.callee.type === "MemberExpression") _out += "(";else _out += ".call(" + SCOPE_LOOKUP + "()" + (tree.arguments.length ? ", " : "");
						_out += tree.arguments.map(_render).join(", ") + ")";
						return _out;
					}

				case "UnaryExpression":
					{
						// single operators: ! - ++ --
						var _out2 = _render(tree.argument);
						if (tree.prefix) _out2 = tree.operator + _out2;else _out2 += tree.operator;
						return _out2;
					}

				case "LogicalExpression":
				case "BinaryExpression":
					{
						// math
						return "(" + _render(tree.left) + tree.operator + _render(tree.right) + ")";
					}

				case "ArrayExpression":
					{
						// arrays
						return "[" + tree.elements.map(_render).join(",") + "]";
					}

				case "ConditionalExpression":
					{
						// if ? then : else
						return "(" + _render(tree.test) + "?" + _render(tree.consequent) + ":" + _render(tree.alternate) + ")";
					}

				default:
					throw new Error("Not supported.");
			}
		}
	}]);
	return Expression;
}(ASTNode);

var Literal = function (_Node) {
	babelHelpers.inherits(Literal, _Node);

	function Literal() {
		babelHelpers.classCallCheck(this, Literal);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Literal).apply(this, arguments));
	}

	babelHelpers.createClass(Literal, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.push(JSON.stringify(this.value));
			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Literal;
}(ASTNode);

var Set = function (_Node) {
	babelHelpers.inherits(Set, _Node);

	function Set() {
		babelHelpers.classCallCheck(this, Set);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Set).apply(this, arguments));
	}

	babelHelpers.createClass(Set, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);
			var safevar = JSON.stringify(this.variable);
			if (this.attribute) this.push("(");else this.push(this.tabs());
			this.push(["scope.set(" + safevar + ", ", exp, ")"]);
			if (!this.attribute) this.push(";\n");else this.push(", \"\")");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Set;
}(ASTNode);

var Raw = function (_Node) {
	babelHelpers.inherits(Raw, _Node);

	function Raw() {
		babelHelpers.classCallCheck(this, Raw);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Raw).apply(this, arguments));
	}

	babelHelpers.createClass(Raw, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var result = void 0;
			var parsers = [].concat((data.tags || {})[this.tagname], this[this.tagname], this.defaultParser).filter(Boolean);

			var attrs = this.attributes.reduce(function (m, a) {
				switch (a.type) {
					case "empty":
						m[a.name] = true;
						break;

					case "string":
						m[a.name] = a.value;
						break;
				}

				return m;
			}, {});

			while (parsers.length && result == null) {
				var parser = parsers.shift();
				if (typeof parser === "function") {
					result = parser.call(this, this.value, attrs, data);
				}
			}

			if (result != null) this.write(result);
			return this.end();
		}
	}, {
		key: "defaultParser",
		value: function defaultParser(value, attrs, data) {
			this._normalize_indent();
			var line = this._line || 1;
			var column = this._column || 1;

			return value.split(/\r?\n/g).map(function (l, i) {
				return new sourceMap.SourceNode(i + line, i === 0 ? column : 1, data.filename, l + "\n");
			});
		}
	}, {
		key: "script",
		value: function script(value, attrs) {
			if (!attrs.src) return;

			var attrsrc = _map(attrs, function (v, n) {
				return "script[" + JSON.stringify(n) + "] = " + JSON.stringify(v) + ";";
			}).join("\n");

			return "(function(){\n\tvar script = document.createElement(\"script\");\n\t" + attrsrc + "\n\tdocument.head.appendChild(script);\n}());\n";
		}
	}, {
		key: "style",
		value: function style(value) {
			return "(function() {\n\tvar style = document.createElement(\"style\");\n\tstyle.innerHTML = " + JSON.stringify(value) + ";\n\tdocument.head.appendChild(style);\n}());";
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Raw;
}(ASTNode);



var AST = Object.freeze({
	utils: utils$1,
	Node: ASTNode,
	File: File,
	Include: Include,
	Comment: Comment,
	Template: Template$1,
	Raw: Raw,
	Element: Element$1,
	Text: Text,
	Attribute: Attribute,
	If: If,
	Branch: Branch,
	Each: Each,
	With: With,
	Set: Set,
	Render: Render,
	Interpolator: Interpolator,
	Expression: Expression,
	Literal: Literal
});

var name = "templejs";
var description = "A JavaScript view framework.";
var author = "Beneath the Ink <info@beneaththeink.com>";
var contributors = [{"name":"Tyler Johnson","email":"tyler@beneaththeink.com","url":"http://github.com/tyler-johnson"}];
var repository = {"type":"git","url":"https://github.com/BeneathTheInk/Temple.git"};
var scripts = {"lint":"eslint src/ test/","test":"make test","build":"make build","prepublish":"make clean && make","autorelease":"autorelease pre && npm publish && autorelease post"};
var main = "lib/index.js";
var bin = "lib/cli.js";
var browser = {"fs-promise":false};
var dependencies = {"assign-props":"^1.1.2","backbone-extend-standalone":"^0.1.2","chalk":"^1.1.1","chokidar":"^1.4.2","dom-matches":"^2.0.0","express":"^4.13.4","fs-promise":"^0.4.1","jsep":"^0.3.0","lodash":"4.x","minimist":"^1.2.0","open":"0.0.5","raf":"^3.1.0","rollup":"^0.25.4","@mrgalaxy/rollup-plugin-inject":"^2.0.0","source-map":"^0.5.3","trackr":"^2.0.2","trackr-objects":"^0.4.0"};
var devDependencies = {"ace-editor-builds":"^1.2.0","autorelease":"^1.6.0","autorelease-github":"^1.1.1","autorelease-travis":"^1.2.1","babel-eslint":"^6.0.0","babel-plugin-lodash":"^2.0.1","babel-preset-es2015-rollup":"^1.1.1","browserify":"^13.0.0","eslint":"~2.2.0","incremental-dom":"^0.3.0","istanbul":"^0.4.2","istanbul-coveralls":"^1.0.3","pegjs":"^0.9.0","random-port":"^0.1.0","rollup-plugin-babel":"^2.3.9","rollup-plugin-commonjs":"^2.2.1","rollup-plugin-json":"^2.0.0","rollup-plugin-node-builtins":"^1.0.0","rollup-plugin-node-globals":"1.0.3","rollup-plugin-node-resolve":"^1.4.0","rollup-plugin-replace":"^1.1.0","tape":"^4.4.0","tape-run":"^2.1.3","uglify-js":"^2.6.1"};
var license = "MIT";
var keywords = ["view","framework","DOM","html","binding","reactive","dependency","mustache"];
var version$1 = "1.0.0";
var pkg = {
	name: name,
	description: description,
	author: author,
	contributors: contributors,
	repository: repository,
	scripts: scripts,
	main: main,
	bin: bin,
	browser: browser,
	dependencies: dependencies,
	devDependencies: devDependencies,
	license: license,
	keywords: keywords,
	version: version$1,
	"jsnext:main": "lib/es6.js"
};

var parser = function () {
  "use strict";

  /*
   * Generated by PEG.js 0.9.0.
   *
   * http://pegjs.org/
   */

  function peg$subclass(child, parent) {
    function ctor() {
      this.constructor = child;
    }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor();
  }

  function peg$SyntaxError(message, expected, found, location) {
    this.message = message;
    this.expected = expected;
    this.found = found;
    this.location = location;
    this.name = "SyntaxError";

    if (typeof Error.captureStackTrace === "function") {
      Error.captureStackTrace(this, peg$SyntaxError);
    }
  }

  peg$subclass(peg$SyntaxError, Error);

  function peg$parse(input) {
    var options = arguments.length > 1 ? arguments[1] : {},
        parser = this,
        peg$FAILED = {},
        peg$startRuleIndices = { start: 0 },
        peg$startRuleIndex = 0,
        peg$consts = [function (nodes) {
      return createNode("File", {
        filename: options.filename,
        source: text(),
        children: _map(nodes, 0)
      });
    }, { type: "other", description: "template" }, "<", { type: "literal", value: "<", description: "\"<\"" }, "template", { type: "literal", value: "template", description: "\"template\"" }, ">", { type: "literal", value: ">", description: "\">\"" }, "</template>", { type: "literal", value: "</template>", description: "\"</template>\"" }, function (attrs, nodes) {
      var name,
          type,
          plugins = [];

      attrs = attrs.filter(function (a) {
        switch (a.name) {
          case "name":
            name = a.value;
            break;

          case "type":
            type = a.value;
            break;

          case "use":
            plugins.push(a.value);
            break;

          default:
            return true;
        }
      });

      return createNode("Template", {
        name: name,
        type: type,
        children: nodes,
        plugins: plugins,
        attributes: attrs
      });
    }, { type: "other", description: "include" }, "include", { type: "literal", value: "include", description: "\"include\"" }, "/>", { type: "literal", value: "/>", description: "\"/>\"" }, function (attrs) {
      var src;

      attrs = attrs.filter(function (a) {
        switch (a.name) {
          case "src":
            src = a.value;
            break;

          default:
            return true;
        }
      });

      return createNode("Include", {
        src: src,
        attributes: attrs
      });
    }, { type: "other", description: "interpolator" }, "{-", { type: "literal", value: "{-", description: "\"{-\"" }, "-}", { type: "literal", value: "-}", description: "\"-}\"" }, { type: "any", description: "any character" }, function (v) {
      return v;
    }, "{{", { type: "literal", value: "{{", description: "\"{{\"" }, "}}", { type: "literal", value: "}}", description: "\"}}\"" }, function (value) {
      return createExpression(value);
    }, function (expression) {
      return createNode("Interpolator", {
        expression: expression
      });
    }, function (nodes) {
      return combineText(nodes);
    }, /^[a-z0-9\-$]/i, { type: "class", value: "[a-z0-9\\-$]i", description: "[a-z0-9\\-$]i" }, function (k) {
      return k.toLowerCase();
    }, "</", { type: "literal", value: "</", description: "\"</\"" }, "{%", { type: "literal", value: "{%", description: "\"{%\"" }, { type: "other", description: "comment" }, "<!--", { type: "literal", value: "<!--", description: "\"<!--\"" }, "-->", { type: "literal", value: "-->", description: "\"-->\"" }, "{#", { type: "literal", value: "{#", description: "\"{#\"" }, "#}", { type: "literal", value: "#}", description: "\"#}\"" }, function (value) {
      return createNode("Comment", { value: value });
    }, { type: "other", description: "raw tag" }, function (tagname) {
      if (tagname[0] === "$") tagname = tagname.substr(1);
      currentRawTag = tagname;
      return _includes(rawTags, tagname);
    }, function (tagname, attrs, v) {
      currentRawTag = null;
      return true;
    }, function (tagname, attrs, v) {
      if (tagname[0] === "$") tagname = tagname.substr(1);

      return createNode("Raw", {
        tagname: tagname,
        attributes: attrs,
        value: v
      });
    }, function (t) {
      if (t[0] === "$") t = t.substr(1);
      return t === currentRawTag;
    }, { type: "other", description: "element" }, function (tagname, attrs) {
      return createNode("Element", {
        tagname: tagname,
        attributes: attrs,
        children: []
      });
    }, function (starttag, attrs, nodes, t) {
      if (starttag !== t) {
        throw expected("</" + starttag + ">");
      }

      return t;
    }, function (starttag, attrs, nodes) {
      return createNode("Element", {
        tagname: starttag,
        attributes: attrs,
        children: nodes
      });
    }, { type: "other", description: "attribute" }, "=", { type: "literal", value: "=", description: "\"=\"" }, function (key) {
      enterAttribute();
      return true;
    }, function (key, v) {
      return v;
    }, function (key, value) {
      exitAttribute();
      return true;
    }, function (key, value) {
      return createNode("Attribute", _assign({
        name: key,
        children: [],
        type: "empty"
      }, value));
    }, "{", { type: "literal", value: "{", description: "\"{\"" }, "}", { type: "literal", value: "}", description: "\"}\"" }, function (exp) {
      var e = jsep(exp);
      var exps = e.type === "Compound" ? e.body : [e];

      return {
        type: "expression",
        children: exps.map(function (e) {
          return createNode("Expression", { tree: e });
        })
      };
    }, "\"", { type: "literal", value: "\"", description: "\"\\\"\"" }, function (n) {
      return {
        children: combineText(n, "Literal"),
        type: "string"
      };
    }, "'", { type: "literal", value: "'", description: "\"'\"" }, function (children) {
      var t = text();
      t = t.substr(1, t.length - 2);

      return _assign({
        value: t
      }, children);
    }, { type: "other", description: "section" }, function (name) {
      if (!_includes(["render", "if", "endif", "each", "endeach", "with", "endwith", "set"], name)) {
        expected("valid template section name");
      }
    }, function (v) {
      return v;
    }, "%}", { type: "literal", value: "%}", description: "\"%}\"" }, function () {
      return attributeMode;
    }, function (n) {
      return combineText(n, "Literal");
    }, { type: "other", description: "if section" }, "if", { type: "literal", value: "if", description: "\"if\"" }, "else", { type: "literal", value: "else", description: "\"else\"" }, function (exp, nodes) {
      return [exp, nodes];
    }, function (exp, nodes, elsifs) {
      return nodes;
    }, "endif", { type: "literal", value: "endif", description: "\"endif\"" }, function (exp, nodes, elsifs, els) {
      var branches = [createNode("Branch", {
        expression: createExpression(exp),
        children: nodes
      })];

      elsifs.forEach(function (b) {
        branches.push(createNode("Branch", {
          expression: createExpression(b[0]),
          children: b[1]
        }));
      });

      if (els) branches.push(createNode("Branch", {
        expression: null,
        children: els
      }));

      return createNode("If", {
        children: branches,
        attribute: attributeMode
      });
    }, { type: "other", description: "each section" }, "each", { type: "literal", value: "each", description: "\"each\"" }, ",", { type: "literal", value: ",", description: "\",\"" }, "in", { type: "literal", value: "in", description: "\"in\"" }, function (l, r) {
      return [l].concat(_map(r, 3));
    }, "endeach", { type: "literal", value: "endeach", description: "\"endeach\"" }, function (vars, exp, nodes) {
      return createNode("Each", {
        expression: createExpression(exp),
        children: nodes,
        variables: vars || [],
        attribute: attributeMode
      });
    }, /^[a-z_$]/i, { type: "class", value: "[a-z_$]i", description: "[a-z_$]i" }, /^[a-z0-9_$]/i, { type: "class", value: "[a-z0-9_$]i", description: "[a-z0-9_$]i" }, { type: "other", description: "render section" }, function () {
      return !attributeMode;
    }, "render", { type: "literal", value: "render", description: "\"render\"" }, function (exp) {
      return createNode("Render", {
        expression: createExpression(exp)
      });
    }, { type: "other", description: "with section" }, "with", { type: "literal", value: "with", description: "\"with\"" }, "endwith", { type: "literal", value: "endwith", description: "\"endwith\"" }, function (exp, nodes) {
      return createNode("With", {
        expression: createExpression(exp),
        children: nodes,
        attribute: attributeMode
      });
    }, { type: "other", description: "set section" }, "set", { type: "literal", value: "set", description: "\"set\"" }, /^[a-z]/i, { type: "class", value: "[a-z]i", description: "[a-z]i" }, function (v, exp) {
      return createNode("Set", {
        variable: v,
        expression: createExpression(exp),
        attribute: attributeMode
      });
    }, "true", { type: "literal", value: "true", description: "\"true\"" }, function () {
      return true;
    }, "false", { type: "literal", value: "false", description: "\"false\"" }, function () {
      return false;
    }, "-", { type: "literal", value: "-", description: "\"-\"" }, /^[0-9]/, { type: "class", value: "[0-9]", description: "[0-9]" }, ".", { type: "literal", value: ".", description: "\".\"" }, function () {
      return parseFloat(text(), 10);
    }, function () {
      return parseInt(text(), 10);
    }, /^[^"]/, { type: "class", value: "[^\"]", description: "[^\"]" }, function (v) {
      return v.join("");
    }, /^[^']/, { type: "class", value: "[^']", description: "[^']" }, "null", { type: "literal", value: "null", description: "\"null\"" }, function () {
      return null;
    }, "undefined", { type: "literal", value: "undefined", description: "\"undefined\"" }, "void", { type: "literal", value: "void", description: "\"void\"" }, /^[,; \t\n\r]/, { type: "class", value: "[,; \\t\\n\\r]", description: "[,; \\t\\n\\r]" }, function () {
      return void 0;
    }, { type: "other", description: "whitespace" }, /^[ \t\n\r]/, { type: "class", value: "[ \\t\\n\\r]", description: "[ \\t\\n\\r]" }, { type: "other", description: "guaranteed whitespace" }, "\\", { type: "literal", value: "\\", description: "\"\\\\\"" }, function (char) {
      return char;
    }],
        peg$bytecode = [peg$decode("%;B/\x82#$%;)./ &;\".) &;!.# &;(/,#;B/#$+\")(\"'#&'#0H*%;)./ &;\".) &;!.# &;(/,#;B/#$+\")(\"'#&'#&/($8\": \"! )(\"'#&'#"), peg$decode("<%2\"\"\"6\"7#/q#;B/h$2$\"\"6$7%/Y$;,/P$2&\"\"6&7'/A$;%/8$2(\"\"6(7)/)$8':*'\"#!)(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7!"), peg$decode("<%2\"\"\"6\"7#/X#;B/O$2,\"\"6,7-/@$;,/7$2.\"\"6.7//($8%:0%!!)(%'#($'#(#'#(\"'#&'#=.\" 7+"), peg$decode("<%%22\"\"6273/\xA2#%$%%<24\"\"6475=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0M*%%<24\"\"6475=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/7$24\"\"6475/($8#:7#!!)(#'#(\"'#&'#.\xB2 &%28\"\"6879/\xA2#%$%%<2:\"\"6:7;=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0M*%%<2:\"\"6:7;=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/7$2:\"\"6:7;/($8#:7#!!)(#'#(\"'#&'#/' 8!:<!! )=.\" 71"), peg$decode("%;#/' 8!:=!! )"), peg$decode("%$;(.5 &;1./ &;$.) &;+.# &;'0;*;(.5 &;1./ &;$.) &;+.# &;'&/' 8!:>!! )"), peg$decode("%;B/Z#%$4?\"\"5!7@/,#0)*4?\"\"5!7@&&&#/\"!&,)/1$;B/($8#:A#!!)(#'#(\"'#&'#"), peg$decode("%%%<%2B\"\"6B7C/;#;&/2$2&\"\"6&7'/#$+#)(#'#(\"'#&'#=.##&&!&'#/g#%<2D\"\"6D7E=.##&&!&'#/L$%<28\"\"6879=.##&&!&'#/1$1\"\"5!76/#$+$)($'#(#'#(\"'#&'#/\"!&,)"), peg$decode("<%%2G\"\"6G7H/\xA2#%$%%<2I\"\"6I7J=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0M*%%<2I\"\"6I7J=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/7$2I\"\"6I7J/($8#:7#!!)(#'#(\"'#&'#.\xB2 &%2K\"\"6K7L/\xA2#%$%%<2M\"\"6M7N=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0M*%%<2M\"\"6M7N=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/7$2M\"\"6M7N/($8#:7#!!)(#'#(\"'#&'#/' 8!:O!! )=.\" 7F"), peg$decode("<%2\"\"\"6\"7#/\xDD#;&/\xD4$9:Q ! -\"\"&!&#/\xC0$;,/\xB7$2&\"\"6&7'/\xA8$%$%%<;*=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0G*%%<;*=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/I$;*/@$9:R #%#!-\"\"&!&#/*$8(:S(#&$\")(('#(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7P"), peg$decode("%2B\"\"6B7C/O#;&/F$9:T ! -\"\"&!&#/2$2&\"\"6&7'/#$+$)($'#(#'#(\"'#&'#"), peg$decode("<%2\"\"\"6\"7#/J#;&/A$;,/8$2.\"\"6.7//)$8$:V$\"\"!)($'#(#'#(\"'#&'#.\x9D &%2\"\"\"6\"7#/\x8D#;&/\x84$;,/{$2&\"\"6&7'/l$;%/c$%2B\"\"6B7C/C#;&/:$2&\"\"6&7'/+$8#:W#$'&$!)(#'#(\"'#&'#/*$8&:X&#$#!)(&'#(%'#($'#(#'#(\"'#&'#=.\" 7U"), peg$decode("$;-0#*;-&"), peg$decode("<%;./\x91#%2Z\"\"6Z7[/X#;B/O$9:\\ !#-\"\"&!&#/;$;0/2$;B/)$8%:]%\"&!)(%'#($'#(#'#(\"'#&'#.\" &\"/>$9:^ \"! -\"\"&!&#/)$8#:_#\"\"!)(#'#(\"'#&'#=.\" 7Y"), peg$decode("%;B/Z#%$4?\"\"5!7@/,#0)*4?\"\"5!7@&&&#/\"!&,)/1$;B/($8#:A#!!)(#'#(\"'#&'#"), peg$decode(";D.) &;1.# &;#"), peg$decode("%%2`\"\"6`7a/¢#%$%%<2b\"\"6b7c=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#0M*%%<2b\"\"6b7c=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&/\"!&,)/7$2b\"\"6b7c/($8#:d#!!)(#'#(\"'#&'#.ŭ &%2e\"\"6e7f/µ#$;/.T &%%%<2e\"\"6e7f=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)0Z*;/.T &%%%<2e\"\"6e7f=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)&/7$2e\"\"6e7f/($8#:g#!!)(#'#(\"'#&'#.Å &%2h\"\"6h7i/µ#$;/.T &%%%<2h\"\"6h7i=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)0Z*;/.T &%%%<2h\"\"6h7i=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)&/7$2h\"\"6h7i/($8#:g#!!)(#'#(\"'#&'#/' 8!:j!! )"), peg$decode("<%%<%;2/8#%;7/' 8!:l!! )/#$+\")(\"'#&'#=/##&'!&&#/I#;5.5 &;6./ &;8.) &;9.# &;:/($8\":m\"! )(\"'#&'#=.\" 7k"), peg$decode("%2D\"\"6D7E/,#;B/#$+\")(\"'#&'#"), peg$decode("%;B/2#2n\"\"6n7o/#$+\")(\"'#&'#"), peg$decode("%9:p  -\"\"&!&#/\x9A#$;/.N &%%%<;2=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)0T*;/.N &%%%<;2=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/\"!&,)&/($8\":q\"! )(\"'#&'#.# &;%"), peg$decode("<%;2/Ư#3s\"\"5\"7t/Ơ$;B/Ɨ$%$;;0#*;;&/\"!&,)/ƀ$;3/ŷ$;4/Ů$$%;2/#3u\"\"5$7v/s$;C/j$3s\"\"5\"7t/[$;B/R$%$;;0#*;;&/\"!&,)/;$;3/2$;4/)$8(:w(\"\" )(('#(''#(&'#(%'#($'#(#'#(\"'#&'#0*%;2/#3u\"\"5$7v/s$;C/j$3s\"\"5\"7t/[$;B/R$%$;;0#*;;&/\"!&,)/;$;3/2$;4/)$8(:w(\"\" )(('#(''#(&'#(%'#($'#(#'#(\"'#&'#&/$%;2/K#3u\"\"5$7v/<$;3/3$;4/*$8$:x$#( %)($'#(#'#(\"'#&'#.\" &\"/L$;2/C$3y\"\"5%7z/4$;3/+$8+:{+$'%$#)(+'#(*'#()'#(('#(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7r"), peg$decode("<%;2/ı#3}\"\"5$7~/Ģ$;B/ę$%;7/§#$%;B/D#2\"\"67/5$;B/,$;7/#$+$)($'#(#'#(\"'#&'#0N*%;B/D#2\"\"67/5$;B/,$;7/#$+$)($'#(#'#(\"'#&'#&/A$;B/8$2\"\"67/)$8$:$\"#\")($'#(#'#(\"'#&'#.\" &\"/}$;B/t$%$;;0#*;;&/\"!&,)/]$;3/T$;4/K$;2/B$3\"\"5'7/3$;3/*$8+:+#'%#)(+'#(*'#()'#(('#(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7|"), peg$decode("%%4\x87\"\"5!7\x88/?#$4\x89\"\"5!7\x8A0)*4\x89\"\"5!7\x8A&/#$+\")(\"'#&'#/\"!&,)"), peg$decode("<%9:\x8C  -\"\"&!&#/i#;2/`$3\x8D\"\"5&7\x8E/Q$;B/H$%$;;0#*;;&/\"!&,)/1$;3/($8&:\x8F&!!)(&'#(%'#($'#(#'#(\"'#&'#=.\" 7\x8B"), peg$decode("<%;2/\x8B#3\x91\"\"5$7\x92/|$;B/s$%$;;0#*;;&/\"!&,)/\\$;3/S$;4/J$;2/A$3\x93\"\"5'7\x94/2$;3/)$8):\x95)\"%#)()'#(('#(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7\x90"), peg$decode("<%;2/\xC4#3\x97\"\"5#7\x98/\xB5$;B/\xAC$%%4\x99\"\"5!7\x9A/E#$4\x89\"\"5!7\x8A/,#0)*4\x89\"\"5!7\x8A&&&#/#$+\")(\"'#&'#/\"!&,)/j$;B/a$2Z\"\"6Z7[/R$;B/I$%$;;0#*;;&/\"!&,)/2$;3/)$8):\x9B)\"%!)()'#(('#(''#(&'#(%'#($'#(#'#(\"'#&'#=.\" 7\x96"), peg$decode("%%<2n\"\"6n7o=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#"), peg$decode("%2\x9C\"\"6\x9C7\x9D/& 8!:\x9E! ).4 &%2\x9F\"\"6\x9F7\xA0/& 8!:\xA1! )"), peg$decode("%2\xA2\"\"6\xA27\xA3.\" &\"/\x89#$4\xA4\"\"5!7\xA5/,#0)*4\xA4\"\"5!7\xA5&&&#/g$%2\xA6\"\"6\xA67\xA7/E#$4\xA4\"\"5!7\xA5/,#0)*4\xA4\"\"5!7\xA5&&&#/#$+\")(\"'#&'#.\" &\"/'$8#:\xA8# )(#'#(\"'#&'#"), peg$decode("%$4\xA4\"\"5!7\xA5/,#0)*4\xA4\"\"5!7\xA5&&&#/& 8!:\xA9! )"), peg$decode("%2e\"\"6e7f/_#$;D.) &4\xAA\"\"5!7\xAB0/*;D.) &4\xAA\"\"5!7\xAB&/7$2e\"\"6e7f/($8#:\xAC#!!)(#'#(\"'#&'#.o &%2h\"\"6h7i/_#$;D.) &4\xAD\"\"5!7\xAE0/*;D.) &4\xAD\"\"5!7\xAE&/7$2h\"\"6h7i/($8#:\xAC#!!)(#'#(\"'#&'#"), peg$decode("%2\xAF\"\"6\xAF7\xB0/& 8!:\xB1! )"), peg$decode("%2\xB2\"\"6\xB27\xB3.\xA6 &%2\xB4\"\"6\xB47\xB5/\x96#;C/\x8D$$%%<4\xB6\"\"5!7\xB7=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#/P#0M*%%<4\xB6\"\"5!7\xB7=.##&&!&'#/1#1\"\"5!76/#$+\")(\"'#&'#&&&#/#$+#)(#'#(\"'#&'#/& 8!:\xB8! )"), peg$decode("<%$4\xBA\"\"5!7\xBB0)*4\xBA\"\"5!7\xBB&/\"!&,)=.\" 7\xB9"), peg$decode("<%$4\xBA\"\"5!7\xBB/,#0)*4\xBA\"\"5!7\xBB&&&#/\"!&,)=.\" 7\xBC"), peg$decode("%2\xBD\"\"6\xBD7\xBE/6#1\"\"5!76/($8\":\xBF\"! )(\"'#&'#")],
        peg$currPos = 0,
        peg$savedPos = 0,
        peg$posDetailsCache = [{ line: 1, column: 1, seenCR: false }],
        peg$maxFailPos = 0,
        peg$maxFailExpected = [],
        peg$silentFails = 0,
        peg$result;

    if ("startRule" in options) {
      if (!(options.startRule in peg$startRuleIndices)) {
        throw new Error("Can't start parsing from rule \"" + options.startRule + "\".");
      }

      peg$startRuleIndex = peg$startRuleIndices[options.startRule];
    }

    function text() {
      return input.substring(peg$savedPos, peg$currPos);
    }

    function location() {
      return peg$computeLocation(peg$savedPos, peg$currPos);
    }

    function expected(description) {
      throw peg$buildException(null, [{ type: "other", description: description }], input.substring(peg$savedPos, peg$currPos), peg$computeLocation(peg$savedPos, peg$currPos));
    }

    function error(message) {
      throw peg$buildException(message, null, input.substring(peg$savedPos, peg$currPos), peg$computeLocation(peg$savedPos, peg$currPos));
    }

    function peg$computePosDetails(pos) {
      var details = peg$posDetailsCache[pos],
          p,
          ch;

      if (details) {
        return details;
      } else {
        p = pos - 1;
        while (!peg$posDetailsCache[p]) {
          p--;
        }

        details = peg$posDetailsCache[p];
        details = {
          line: details.line,
          column: details.column,
          seenCR: details.seenCR
        };

        while (p < pos) {
          ch = input.charAt(p);
          if (ch === "\n") {
            if (!details.seenCR) {
              details.line++;
            }
            details.column = 1;
            details.seenCR = false;
          } else if (ch === "\r" || ch === "\u2028" || ch === "\u2029") {
            details.line++;
            details.column = 1;
            details.seenCR = true;
          } else {
            details.column++;
            details.seenCR = false;
          }

          p++;
        }

        peg$posDetailsCache[pos] = details;
        return details;
      }
    }

    function peg$computeLocation(startPos, endPos) {
      var startPosDetails = peg$computePosDetails(startPos),
          endPosDetails = peg$computePosDetails(endPos);

      return {
        start: {
          offset: startPos,
          line: startPosDetails.line,
          column: startPosDetails.column
        },
        end: {
          offset: endPos,
          line: endPosDetails.line,
          column: endPosDetails.column
        }
      };
    }

    function peg$fail(expected) {
      if (peg$currPos < peg$maxFailPos) {
        return;
      }

      if (peg$currPos > peg$maxFailPos) {
        peg$maxFailPos = peg$currPos;
        peg$maxFailExpected = [];
      }

      peg$maxFailExpected.push(expected);
    }

    function peg$buildException(message, expected, found, location) {
      function cleanupExpected(expected) {
        var i = 1;

        expected.sort(function (a, b) {
          if (a.description < b.description) {
            return -1;
          } else if (a.description > b.description) {
            return 1;
          } else {
            return 0;
          }
        });

        while (i < expected.length) {
          if (expected[i - 1] === expected[i]) {
            expected.splice(i, 1);
          } else {
            i++;
          }
        }
      }

      function buildMessage(expected, found) {
        function stringEscape(s) {
          function hex(ch) {
            return ch.charCodeAt(0).toString(16).toUpperCase();
          }

          return s.replace(/\\/g, '\\\\').replace(/"/g, '\\"').replace(/\x08/g, '\\b').replace(/\t/g, '\\t').replace(/\n/g, '\\n').replace(/\f/g, '\\f').replace(/\r/g, '\\r').replace(/[\x00-\x07\x0B\x0E\x0F]/g, function (ch) {
            return '\\x0' + hex(ch);
          }).replace(/[\x10-\x1F\x80-\xFF]/g, function (ch) {
            return '\\x' + hex(ch);
          }).replace(/[\u0100-\u0FFF]/g, function (ch) {
            return "\\u0" + hex(ch);
          }).replace(/[\u1000-\uFFFF]/g, function (ch) {
            return "\\u" + hex(ch);
          });
        }

        var expectedDescs = new Array(expected.length),
            expectedDesc,
            foundDesc,
            i;

        for (i = 0; i < expected.length; i++) {
          expectedDescs[i] = expected[i].description;
        }

        expectedDesc = expected.length > 1 ? expectedDescs.slice(0, -1).join(", ") + " or " + expectedDescs[expected.length - 1] : expectedDescs[0];

        foundDesc = found ? "\"" + stringEscape(found) + "\"" : "end of input";

        return "Expected " + expectedDesc + " but " + foundDesc + " found.";
      }

      if (expected !== null) {
        cleanupExpected(expected);
      }

      return new peg$SyntaxError(message !== null ? message : buildMessage(expected, found), expected, found, location);
    }

    function peg$decode(s) {
      var bc = new Array(s.length),
          i;

      for (i = 0; i < s.length; i++) {
        bc[i] = s.charCodeAt(i) - 32;
      }

      return bc;
    }

    function peg$parseRule(index) {
      var bc = peg$bytecode[index],
          ip = 0,
          ips = [],
          end = bc.length,
          ends = [],
          stack = [],
          params,
          i;

      while (true) {
        while (ip < end) {
          switch (bc[ip]) {
            case 0:
              stack.push(peg$consts[bc[ip + 1]]);
              ip += 2;
              break;

            case 1:
              stack.push(void 0);
              ip++;
              break;

            case 2:
              stack.push(null);
              ip++;
              break;

            case 3:
              stack.push(peg$FAILED);
              ip++;
              break;

            case 4:
              stack.push([]);
              ip++;
              break;

            case 5:
              stack.push(peg$currPos);
              ip++;
              break;

            case 6:
              stack.pop();
              ip++;
              break;

            case 7:
              peg$currPos = stack.pop();
              ip++;
              break;

            case 8:
              stack.length -= bc[ip + 1];
              ip += 2;
              break;

            case 9:
              stack.splice(-2, 1);
              ip++;
              break;

            case 10:
              stack[stack.length - 2].push(stack.pop());
              ip++;
              break;

            case 11:
              stack.push(stack.splice(stack.length - bc[ip + 1], bc[ip + 1]));
              ip += 2;
              break;

            case 12:
              stack.push(input.substring(stack.pop(), peg$currPos));
              ip++;
              break;

            case 13:
              ends.push(end);
              ips.push(ip + 3 + bc[ip + 1] + bc[ip + 2]);

              if (stack[stack.length - 1]) {
                end = ip + 3 + bc[ip + 1];
                ip += 3;
              } else {
                end = ip + 3 + bc[ip + 1] + bc[ip + 2];
                ip += 3 + bc[ip + 1];
              }

              break;

            case 14:
              ends.push(end);
              ips.push(ip + 3 + bc[ip + 1] + bc[ip + 2]);

              if (stack[stack.length - 1] === peg$FAILED) {
                end = ip + 3 + bc[ip + 1];
                ip += 3;
              } else {
                end = ip + 3 + bc[ip + 1] + bc[ip + 2];
                ip += 3 + bc[ip + 1];
              }

              break;

            case 15:
              ends.push(end);
              ips.push(ip + 3 + bc[ip + 1] + bc[ip + 2]);

              if (stack[stack.length - 1] !== peg$FAILED) {
                end = ip + 3 + bc[ip + 1];
                ip += 3;
              } else {
                end = ip + 3 + bc[ip + 1] + bc[ip + 2];
                ip += 3 + bc[ip + 1];
              }

              break;

            case 16:
              if (stack[stack.length - 1] !== peg$FAILED) {
                ends.push(end);
                ips.push(ip);

                end = ip + 2 + bc[ip + 1];
                ip += 2;
              } else {
                ip += 2 + bc[ip + 1];
              }

              break;

            case 17:
              ends.push(end);
              ips.push(ip + 3 + bc[ip + 1] + bc[ip + 2]);

              if (input.length > peg$currPos) {
                end = ip + 3 + bc[ip + 1];
                ip += 3;
              } else {
                end = ip + 3 + bc[ip + 1] + bc[ip + 2];
                ip += 3 + bc[ip + 1];
              }

              break;

            case 18:
              ends.push(end);
              ips.push(ip + 4 + bc[ip + 2] + bc[ip + 3]);

              if (input.substr(peg$currPos, peg$consts[bc[ip + 1]].length) === peg$consts[bc[ip + 1]]) {
                end = ip + 4 + bc[ip + 2];
                ip += 4;
              } else {
                end = ip + 4 + bc[ip + 2] + bc[ip + 3];
                ip += 4 + bc[ip + 2];
              }

              break;

            case 19:
              ends.push(end);
              ips.push(ip + 4 + bc[ip + 2] + bc[ip + 3]);

              if (input.substr(peg$currPos, peg$consts[bc[ip + 1]].length).toLowerCase() === peg$consts[bc[ip + 1]]) {
                end = ip + 4 + bc[ip + 2];
                ip += 4;
              } else {
                end = ip + 4 + bc[ip + 2] + bc[ip + 3];
                ip += 4 + bc[ip + 2];
              }

              break;

            case 20:
              ends.push(end);
              ips.push(ip + 4 + bc[ip + 2] + bc[ip + 3]);

              if (peg$consts[bc[ip + 1]].test(input.charAt(peg$currPos))) {
                end = ip + 4 + bc[ip + 2];
                ip += 4;
              } else {
                end = ip + 4 + bc[ip + 2] + bc[ip + 3];
                ip += 4 + bc[ip + 2];
              }

              break;

            case 21:
              stack.push(input.substr(peg$currPos, bc[ip + 1]));
              peg$currPos += bc[ip + 1];
              ip += 2;
              break;

            case 22:
              stack.push(peg$consts[bc[ip + 1]]);
              peg$currPos += peg$consts[bc[ip + 1]].length;
              ip += 2;
              break;

            case 23:
              stack.push(peg$FAILED);
              if (peg$silentFails === 0) {
                peg$fail(peg$consts[bc[ip + 1]]);
              }
              ip += 2;
              break;

            case 24:
              peg$savedPos = stack[stack.length - 1 - bc[ip + 1]];
              ip += 2;
              break;

            case 25:
              peg$savedPos = peg$currPos;
              ip++;
              break;

            case 26:
              params = bc.slice(ip + 4, ip + 4 + bc[ip + 3]);
              for (i = 0; i < bc[ip + 3]; i++) {
                params[i] = stack[stack.length - 1 - params[i]];
              }

              stack.splice(stack.length - bc[ip + 2], bc[ip + 2], peg$consts[bc[ip + 1]].apply(null, params));

              ip += 4 + bc[ip + 3];
              break;

            case 27:
              stack.push(peg$parseRule(bc[ip + 1]));
              ip += 2;
              break;

            case 28:
              peg$silentFails++;
              ip++;
              break;

            case 29:
              peg$silentFails--;
              ip++;
              break;

            default:
              throw new Error("Invalid opcode: " + bc[ip] + ".");
          }
        }

        if (ends.length > 0) {
          end = ends.pop();
          ip = ips.pop();
        } else {
          break;
        }
      }

      return stack[0];
    }

    options = _assign({
      strict: true
    }, options);

    function createNode(type, props) {
      var loc = location();
      if (typeof type !== "function") type = AST[type];
      return new type(loc.start.line, loc.start.column, props);
    }

    function combineText(nodes, type) {
      return nodes.reduce(function (m, n) {
        var last = m.length - 1;
        if (typeof n === "string" && typeof m[last] === "string") {
          m[last] += n;
        } else {
          m.push(n);
        }
        return m;
      }, []).map(function (n) {
        if (typeof n !== "string") return n;
        return createNode(type || "Text", { value: n });
      });
    }

    function createExpression(v) {
      return createNode("Expression", {
        value: v,
        tree: jsep(v)
      });
    }

    var attributeMode = false;
    function enterAttribute() {
      attributeMode = true;
    }
    function exitAttribute() {
      attributeMode = false;
    }

    var rawTags = _union(["script", "style"], Object.keys(options.tags || {}));
    var currentRawTag;

    peg$result = peg$parseRule(peg$startRuleIndex);

    if (peg$result !== peg$FAILED && peg$currPos === input.length) {
      return peg$result;
    } else {
      if (peg$result !== peg$FAILED && peg$currPos < input.length) {
        peg$fail({ type: "end", description: "end of input" });
      }

      throw peg$buildException(null, peg$maxFailExpected, peg$maxFailPos < input.length ? input.charAt(peg$maxFailPos) : null, peg$maxFailPos < input.length ? peg$computeLocation(peg$maxFailPos, peg$maxFailPos + 1) : peg$computeLocation(peg$maxFailPos, peg$maxFailPos));
    }
  }

  return {
    SyntaxError: peg$SyntaxError,
    parse: peg$parse
  };
}();
var parse$1 = parser.parse;

var smfurl = "sourceMappingURL=";
var datauri = "data:application/json;charset=utf-8;base64,";
var rel_regex = /^\.{0,2}\//;

var toBase64;
if (typeof window !== "undefined" && typeof window.btoa === "function") {
	toBase64 = window.btoa;
} else toBase64 = function toBase64(str) {
	return new Buffer(str, "utf-8").toString("base64");
};

function parseFiles(files, _ref) {
	var filename = _ref.filename;

	if (typeof files === "string") {
		if (filename) files = babelHelpers.defineProperty({}, filename, files);else files = [files];
	}

	if (Array.isArray(files)) files = files.reduce(function (memo, src, name) {
		if (typeof name === "number") name = "template" + name + ".html";
		memo[name] = src;
		return memo;
	}, {});

	return files;
}

function parse(src) {
	var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

	try {
		return parse$1(src, options);
	} catch (e) {
		e.filename = options.filename;
		e.source = src;
		throw e;
	}
}

function srcToString(smf) {
	return this.code + "\n\n//# " + smfurl + (typeof smf === "string" ? smf : datauri + toBase64(this.map.toString()));
}

function compile(files) {
	var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
	var cb = arguments[2];

	files = parseFiles(files, options);
	var exts = [].concat(options.extensions || ".html").filter(Boolean);
	var basedir = options.basedir || (options.filename ? path.dirname(options.filename) : ".");
	var hasCb = typeof cb === "function";
	var templates = [];

	return rollup.rollup({
		onwarn: function onwarn() {},
		entry: "_entry.js",
		plugins: [{
			resolveId: function resolveId(id, file) {
				if (id === "_entry.js" || id === "_template.js") return id;

				var full = path.resolve(basedir, path.dirname(file), id);
				var entry = _find(Object.keys(files), function (name) {
					return path.resolve(basedir, name) === full;
				});
				if (entry) return entry;

				// only relative paths are followed
				if (rel_regex.test(id) && _includes(exts, path.extname(id))) {
					return full;
				}
			},
			load: function load(id) {
				if (id === "_entry.js") {
					return Object.keys(files).map(function (name) {
						return "import " + JSON.stringify(name);
					}).join("\n") + "\nexport default Template;\n";
				}

				if (id === "_template.js") {
					return "export default {};\n";
				}

				if (_has(files, id)) {
					return files[id];
				}
			},
			transform: function transform(src, id) {
				if (!_includes(exts, path.extname(id))) return;
				templates.push(id);

				var file = parse(src, _assign({ filename: id }, options));
				var out = file.compile(options).toStringWithSourceMap();
				out.map = out.map.toJSON();
				return out;
			}
		}, inject({
			Temple: "templejs",
			Template: "_template.js",
			idom: ["templejs", "idom"],
			decorators: ["templejs", "decorators"]
		})].concat(options.plugins).filter(Boolean)
	}).then(function (bundle) {
		var out = bundle.generate({
			format: options.format,
			sourceMap: true,
			exports: "default",
			useStrict: false,
			moduleId: options.moduleId,
			moduleName: options.moduleName,
			globals: {
				templejs: "Temple"
			}
		});

		out.templates = templates;
		out.toString = srcToString;

		return out;
	}).then(function (res) {
		if (hasCb) cb(null, res);
		return res;
	}, function (e) {
		if (hasCb) cb(e);
		throw e;
	});
}

function exec(tpl, options) {
	return compile(tpl, _assign({
		format: "iife",
		moduleName: "Template"
	}, options)).then(function (r) {
		return new Function("Temple", r.code)(Temple);
	});
}

// export function compileHTML(html, options) {
// 	options = options || {};
// 	var node = new AST.HTML(null, html);
// 	return getSource(node.compile(options), html, options);
// }
//
// export function renderHTML(html, options) {
// 	/* jshint -W054 */
// 	var r = compileHTML(html, options);
// 	// console.log(r);
// 	return (new Function("Temple", r))(require("./"));
// }

function Each$1(val, vars, scope, fn, that) {
	return map(val, function (item, key) {
		var nscope = new Scope(scope);

		if (vars.length === 1) nscope.set(vars[0], item);else if (vars.length == 2) {
			nscope.set(vars[0], key);
			nscope.set(vars[1], item);
		}

		return fn.call(that, nscope, key, val);
	});
}

function With$1(val, scope, fn, that) {
	var nscope = new Scope(scope);
	nscope.set(val);
	nscope.dataVar.set(val);
	return fn.call(that, nscope);
}

function Element$2(tagname, key, scope, fn, that) {
	var node = elementOpen(tagname, key);
	var nscope = new Scope(null, scope);
	var comp = autoelement(node, function () {
		return fn.call(that, nscope, node);
	});
	elementClose(tagname);
	return comp;
}

var version = pkg.version || "edge";



var Temple = Object.freeze({
	version: version,
	Trackr: Trackr,
	Scope: Scope,
	AST: AST,
	idom: idom,
	utils: utils,
	plugins: plugins$1,
	decorators: decorators$1,
	actions: actions$1,
	Map: trackrObjects.Map,
	List: trackrObjects.List,
	Variable: trackrObjects.Variable,
	trackProperty: trackrObjects.trackProperty,
	templates: templates,
	types: types,
	Template: Template,
	helpers: helpers,
	getHelper: getHelper,
	render: render,
	getByName: getByName,
	paint: paint,
	registerType: registerType,
	create: create$1,
	parse: parse,
	compile: compile,
	exec: exec,
	Each: Each$1,
	With: With$1,
	Element: Element$2
});

exports.version = version;
exports.Trackr = Trackr;
exports.Scope = Scope;
exports.AST = AST;
exports.idom = idom;
exports.utils = utils;
exports.plugins = plugins$1;
exports.decorators = decorators$1;
exports.actions = actions$1;
exports.Map = trackrObjects.Map;
exports.List = trackrObjects.List;
exports.Variable = trackrObjects.Variable;
exports.trackProperty = trackrObjects.trackProperty;
exports.templates = templates;
exports.types = types;
exports.Template = Template;
exports.helpers = helpers;
exports.getHelper = getHelper;
exports.render = render;
exports.getByName = getByName;
exports.paint = paint;
exports.registerType = registerType;
exports.create = create$1;
exports.parse = parse;
exports.compile = compile;
exports.exec = exec;
exports.Each = Each$1;
exports.With = With$1;
exports.Element = Element$2;